# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""
AWSCLI
----
A Universal Command Line Environment for Amazon Web Services.
"""

import importlib.abc
import os
import sys

__version__ = '2.34.29'

#
# Get our data path to be added to botocore's search path
#
_awscli_data_path = []
if 'AWS_DATA_PATH' in os.environ:
    for path in os.environ['AWS_DATA_PATH'].split(os.pathsep):
        path = os.path.expandvars(path)
        path = os.path.expanduser(path)
        _awscli_data_path.append(path)
_awscli_data_path.append(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
)
os.environ['AWS_DATA_PATH'] = os.pathsep.join(_awscli_data_path)


SCALAR_TYPES = set(
    [
        'string',
        'float',
        'integer',
        'long',
        'boolean',
        'double',
        'blob',
        'timestamp',
    ]
)
COMPLEX_TYPES = set(['structure', 'map', 'list'])


class TopLevelImportAliasFinder(importlib.abc.MetaPathFinder):
    """Finder to allow missing awscli namespace in package imports

    This allows packages that used to be dependencies of the AWS CLI and are
    now maintained in the AWS CLI codebase to still be importable as a
    top-level import and minimize churn on the codebase. For example, both of
    these import statements result in the import of the same module on disk::

        >> import awscli.botocore
        >> import botocore

    Note: That this import alias only comes into affect if anything is
    imported from the awscli package.
    """

    _PACKAGES = [
        'botocore',
        's3transfer',
    ]
    _TARGET_FINDERS = [
        'pyimod02_importers.PyiFrozenImporter',  # Pyinstaller injected finder
        '_frozen_importlib_external.PathFinder',  # Built-in path finder
    ]

    def __init__(self, underlying_finder):
        self._underlying_finder = underlying_finder
        self._parent_path = [os.path.dirname(__file__)]

    @classmethod
    def add_alias_finder(cls, meta_path):
        """Wrap known MetaPathFinders that search for aliased packages"""
        for i, finder in enumerate(meta_path):
            # We are searching by module + name because the meta path finders
            # that we want to add the alias shim for either cannot be imported
            # (e.g. the PyInstaller meta path finder) or is recommended against
            # being directly imported (e.g. the standard library path finder).
            if isinstance(finder, type):
                finder_name = finder.__name__
            else:
                finder_name = finder.__class__.__name__
            full_cls_name = f'{finder.__module__}.{finder_name}'
            if full_cls_name in cls._TARGET_FINDERS:
                meta_path.insert(i, cls(finder))
                return

    def find_spec(self, fullname, path, target=None):
        if fullname in self._PACKAGES:
            path = self._parent_path
        return self._underlying_finder.find_spec(fullname, path, target)


TopLevelImportAliasFinder.add_alias_finder(sys.meta_path)

_DEFAULT_BASE_REMOTE_URL = f"https://awscli.amazonaws.com/v2/documentation/api/{__version__}"   # noqa



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

from botocore import xform_name
from botocore.compat import OrderedDict
from botocore.docs.bcdoc.restdoc import DocumentStructure
from botocore.docs.method import document_model_driven_method
from botocore.docs.utils import DocumentedShape
from botocore.utils import get_service_module_name


class PaginatorDocumenter:
    def __init__(self, client, service_paginator_model, root_docs_path):
        self._client = client
        self._client_class_name = self._client.__class__.__name__
        self._service_name = self._client.meta.service_model.service_name
        self._service_paginator_model = service_paginator_model
        self._root_docs_path = root_docs_path
        self._USER_GUIDE_LINK = (
            'https://boto3.amazonaws.com/'
            'v1/documentation/api/latest/guide/paginators.html'
        )

    def document_paginators(self, section):
        """Documents the various paginators for a service

        param section: The section to write to.
        """
        section.style.h2('Paginators')
        self._add_overview(section)
        section.style.new_line()
        section.writeln('The available paginators are:')
        section.style.toctree()

        paginator_names = sorted(
            self._service_paginator_model._paginator_config
        )

        # List the available paginators and then document each paginator.
        for paginator_name in paginator_names:
            section.style.tocitem(
                f'{self._service_name}/paginator/{paginator_name}'
            )
            # Create a new DocumentStructure for each paginator and add contents.
            paginator_doc_structure = DocumentStructure(
                paginator_name, target='html'
            )
            self._add_paginator(paginator_doc_structure, paginator_name)
            # Write paginators in individual/nested files.
            # Path: <root>/reference/services/<service>/paginator/<paginator_name>.rst
            paginator_dir_path = os.path.join(
                self._root_docs_path, self._service_name, 'paginator'
            )
            paginator_doc_structure.write_to_file(
                paginator_dir_path, paginator_name
            )

    def _add_paginator(self, section, paginator_name):
        section.add_title_section(paginator_name)

        # Docment the paginator class
        paginator_section = section.add_new_section(paginator_name)
        paginator_section.style.start_sphinx_py_class(
            class_name=f'{self._client_class_name}.Paginator.{paginator_name}'
        )
        paginator_section.style.start_codeblock()
        paginator_section.style.new_line()

        # Document how to instantiate the paginator.
        paginator_section.write(
            f'paginator = client.get_paginator(\'{xform_name(paginator_name)}\')'
        )
        paginator_section.style.end_codeblock()
        paginator_section.style.new_line()
        # Get the pagination model for the particular paginator.
        paginator_config = self._service_paginator_model.get_paginator(
            paginator_name
        )
        document_paginate_method(
            section=paginator_section,
            paginator_name=paginator_name,
            event_emitter=self._client.meta.events,
            service_model=self._client.meta.service_model,
            paginator_config=paginator_config,
        )

    def _add_overview(self, section):
        section.style.new_line()
        section.write(
            'Paginators are available on a client instance '
            'via the ``get_paginator`` method. For more detailed instructions '
            'and examples on the usage of paginators, see the '
            'paginators '
        )
        section.style.external_link(
            title='user guide',
            link=self._USER_GUIDE_LINK,
        )
        section.write('.')
        section.style.new_line()


def document_paginate_method(
    section,
    paginator_name,
    event_emitter,
    service_model,
    paginator_config,
    include_signature=True,
):
    """Documents the paginate method of a paginator

    :param section: The section to write to

    :param paginator_name: The name of the paginator. It is snake cased.

    :param event_emitter: The event emitter to use to emit events

    :param service_model: The service model

    :param paginator_config: The paginator config associated to a particular
        paginator.

    :param include_signature: Whether or not to include the signature.
        It is useful for generating docstrings.
    """
    # Retrieve the operation model of the underlying operation.
    operation_model = service_model.operation_model(paginator_name)

    # Add representations of the request and response parameters
    # we want to include in the description of the paginate method.
    # These are parameters we expose via the botocore interface.
    pagination_config_members = OrderedDict()

    pagination_config_members['MaxItems'] = DocumentedShape(
        name='MaxItems',
        type_name='integer',
        documentation=(
            '<p>The total number of items to return. If the total '
            'number of items available is more than the value '
            'specified in max-items then a <code>NextToken</code> '
            'will be provided in the output that you can use to '
            'resume pagination.</p>'
        ),
    )

    if paginator_config.get('limit_key', None):
        pagination_config_members['PageSize'] = DocumentedShape(
            name='PageSize',
            type_name='integer',
            documentation='<p>The size of each page.<p>',
        )

    pagination_config_members['StartingToken'] = DocumentedShape(
        name='StartingToken',
        type_name='string',
        documentation=(
            '<p>A token to specify where to start paginating. '
            'This is the <code>NextToken</code> from a previous '
            'response.</p>'
        ),
    )

    botocore_pagination_params = [
        DocumentedShape(
            name='PaginationConfig',
            type_name='structure',
            documentation=(
                '<p>A dictionary that provides parameters to control '
                'pagination.</p>'
            ),
            members=pagination_config_members,
        )
    ]

    botocore_pagination_response_params = [
        DocumentedShape(
            name='NextToken',
            type_name='string',
            documentation=('<p>A token to resume pagination.</p>'),
        )
    ]

    service_pagination_params = []

    # Add the normal input token of the method to a list
    # of input paramters that we wish to hide since we expose our own.
    if isinstance(paginator_config['input_token'], list):
        service_pagination_params += paginator_config['input_token']
    else:
        service_pagination_params.append(paginator_config['input_token'])

    # Hide the limit key in the documentation.
    if paginator_config.get('limit_key', None):
        service_pagination_params.append(paginator_config['limit_key'])

    # Hide the output tokens in the documentation.
    service_pagination_response_params = []
    if isinstance(paginator_config['output_token'], list):
        service_pagination_response_params += paginator_config['output_token']
    else:
        service_pagination_response_params.append(
            paginator_config['output_token']
        )

    paginate_description = (
        'Creates an iterator that will paginate through responses '
        f'from :py:meth:`{get_service_module_name(service_model)}.Client.{xform_name(paginator_name)}`.'
    )

    document_model_driven_method(
        section,
        'paginate',
        operation_model,
        event_emitter=event_emitter,
        method_description=paginate_description,
        example_prefix='response_iterator = paginator.paginate',
        include_input=botocore_pagination_params,
        include_output=botocore_pagination_response_params,
        exclude_input=service_pagination_params,
        exclude_output=service_pagination_response_params,
        include_signature=include_signature,
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.autocomplete.serverside import servercomp


def add_log_completers(custom_completers):
    custom_completers.append(GroupNameCompleter())


class GroupNameCompleter(servercomp.BaseCustomServerSideCompleter):
    _PARAM_NAME = 'group_name'
    _LINEAGE = ['aws', 'logs']
    _COMMAND_NAMES = ['tail']

    def _get_remote_results(self, parsed):
        client = self._get_client('logs', parsed)
        response = self._invoke_api(client, 'describe_log_groups', {})
        return [
            group['logGroupName'] for group in response.get('logGroups', [])
        ]



# Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import os
import random
import string
import sys

# Import botocore at module level to avoid repeated imports
import botocore.session
from botocore.exceptions import ClientError, ProfileNotFound

from awscli.compat import compat_input
from awscli.customizations.commands import BasicCommand
from awscli.customizations.configure import profile_to_section
from awscli.customizations.configure.writer import ConfigFileWriter
from awscli.errorformat import write_error

LOG = logging.getLogger(__name__)


class InteractiveMFAPrompter:
    """Handles interactive prompting for MFA login."""

    def get_value(self, current_value, prompt_text=''):
        """Prompt for a value, showing the current value as a default."""
        response = compat_input(f"{prompt_text} [{current_value}]: ")
        if not response:
            # If the user hits enter, return the current value
            return current_value
        return response

    def get_credential_value(self, current_value, config_name, prompt_text=''):
        """Prompt for credential values with masking for sensitive data."""
        response = compat_input(f"{prompt_text}: ")
        if not response:
            return None
        return response


class ConfigureMFALoginCommand(BasicCommand):
    """Configures MFA login for AWS CLI by creating temporary credentials."""

    NAME = 'mfa-login'
    DESCRIPTION = BasicCommand.FROM_FILE(
        'configure', 'mfa-login', '_description.rst'
    )
    SYNOPSIS = (
        'aws configure mfa-login [--profile profile-name] '
        '[--update-profile profile-to-update] [--duration-seconds seconds] '
        '[--serial-number mfa-serial-number]'
    )
    EXAMPLES = BasicCommand.FROM_FILE('configure', 'mfa-login', '_examples.rst')
    ARG_TABLE = [
        {
            'name': 'update-profile',
            'help_text': (
                'The profile to update with temporary credentials. '
                'If not provided, a default name will be generated.'
            ),
            'action': 'store',
            'required': False,
            'cli_type_name': 'string',
        },
        {
            'name': 'duration-seconds',
            'help_text': (
                'The duration, in seconds, that the credentials should remain valid. '
                'Minimum is 900 seconds (15 minutes), maximum is 129600 seconds (36 hours).'
            ),
            'action': 'store',
            'required': False,
            'cli_type_name': 'integer',
            'default': 43200,
        },
        {
            'name': 'serial-number',
            'help_text': (
                'The ARN or serial number of the MFA device associated with the IAM user. '
                'If not provided, will use the mfa_serial from the profile configuration.'
            ),
            'action': 'store',
            'required': False,
            'cli_type_name': 'string',
        },
    ]

    # Values to prompt for during interactive setup
    VALUES_TO_PROMPT = [
        ('aws_access_key_id', 'AWS Access Key ID'),
        ('aws_secret_access_key', 'AWS Secret Access Key'),
        ('mfa_serial', 'MFA serial number or ARN'),
        ('mfa_token', 'MFA token code'),
    ]

    def __init__(self, session, prompter=None, config_writer=None):
        super().__init__(session)
        if prompter is None:
            prompter = InteractiveMFAPrompter()
        self._prompter = prompter
        if config_writer is None:
            config_writer = ConfigFileWriter()
        self._config_writer = config_writer

    def _generate_profile_name_from_mfa(self, mfa_serial):
        """Generate a deterministic profile name from MFA serial/ARN."""
        if isinstance(mfa_serial, str) and mfa_serial.startswith('arn:aws:iam::'):
            # Parse ARN: arn:aws:iam::123456789012:mfa/device-name
            parts = mfa_serial.split(':')
            account_id = parts[4]
            device_name = parts[5].split('/')[-1]  # Get device name after 'mfa/'
            return f"{account_id}-{device_name}"
        else:
            # Assume it's just a serial number
            return f"session-{mfa_serial}"

    def _get_target_profile(self, parsed_args, mfa_serial=None):
        """Get or generate the target profile name."""
        target_profile = parsed_args.update_profile
        if not target_profile:
            if mfa_serial:
                target_profile = self._generate_profile_name_from_mfa(mfa_serial)
            else:
                target_profile = "session-temp"
            target_profile = self._prompter.get_value(
                target_profile, 'Profile to update'
            )
        return target_profile

    def _get_mfa_token(self):
        """Prompt for MFA token code."""
        token_code = self._prompter.get_credential_value(
            'None', 'mfa_token', 'MFA token code'
        )
        if not token_code:
            write_error(sys.stderr, "MFA token code is required")
            return None
        return token_code

    def _call_sts_get_session_token(self, sts_client, duration_seconds, mfa_serial, token_code):
        """Call STS to get temporary credentials."""
        try:
            response = sts_client.get_session_token(
                DurationSeconds=duration_seconds,
                SerialNumber=mfa_serial,
                TokenCode=token_code,
            )
            return response
        except ClientError as e:
            write_error(sys.stderr, f"An error occurred: {e}")
            return None

    def _resolve_mfa_serial(self, parsed_args, source_config):
        """Resolve MFA serial from args, config, or prompt."""
        mfa_serial = parsed_args.serial_number or source_config.get('mfa_serial')
        if not mfa_serial:
            mfa_serial = self._prompter.get_credential_value(
                'None', 'mfa_serial', 'MFA serial number or ARN'
            )
            if not mfa_serial:
                write_error(sys.stderr, "MFA serial number or MFA device ARN is required")
                return None
        return mfa_serial

    def _write_temporary_credentials(self, temp_credentials, target_profile):
        """Write temporary credentials to the credentials file."""
        credentials_file = os.path.expanduser(self._session.get_config_variable('credentials_file'))

        credential_values = {
            '__section__': target_profile,
            'aws_access_key_id': temp_credentials['AccessKeyId'],
            'aws_secret_access_key': temp_credentials['SecretAccessKey'],
            'aws_session_token': temp_credentials['SessionToken'],
        }

        try:
            expiration_time = temp_credentials['Expiration'].strftime(
                '%Y-%m-%d %H:%M:%S UTC'
            )
        except AttributeError:
            expiration_time = str(temp_credentials['Expiration'])

        self._config_writer.update_config(credential_values, credentials_file)

        sys.stdout.write(
            f"Temporary credentials written to profile '{target_profile}'\n"
        )
        sys.stdout.write(f"Credentials will expire at {expiration_time}\n")
        sys.stdout.write(
            f"To use these credentials, specify --profile {target_profile} when running AWS CLI commands\n"
        )
        return 0

    def _run_main(self, parsed_args, parsed_globals):
        duration_seconds = parsed_args.duration_seconds

        # Use the CLI session directly
        credentials = self._session.get_credentials()
        if credentials is None:
            return self._handle_interactive_prompting(parsed_args, duration_seconds)
        
        source_config = self._session.get_scoped_config()

        # Resolve MFA serial number
        mfa_serial = self._resolve_mfa_serial(parsed_args, source_config)
        if not mfa_serial:
            return 1

        # Get MFA token code
        token_code = self._get_mfa_token()
        if not token_code:
            return 1

        # Get the target profile name
        target_profile = self._get_target_profile(parsed_args, mfa_serial)

        # Call STS to get temporary credentials
        sts_client = self._session.create_client('sts')
        response = self._call_sts_get_session_token(
            sts_client, duration_seconds, mfa_serial, token_code
        )
        if not response:
            return 1

        # Write credentials and return
        return self._write_temporary_credentials(
            response['Credentials'], target_profile
        )

    def _handle_interactive_prompting(self, parsed_args, duration_seconds):
        """Handle the case where no default profile exists, and there is no profile explicitly named as a configuration source"""
        sys.stdout.write(
            "Please provide your AWS credentials:\n"
        )

        values = {}
        for config_name, prompt_text in self.VALUES_TO_PROMPT:
            if config_name == 'mfa_serial' and parsed_args.serial_number:
                values[config_name] = parsed_args.serial_number
                continue
                
            value = self._prompter.get_credential_value(
                'None', config_name, prompt_text
            )
            if not value or value == 'None':
                write_error(sys.stderr, f"{prompt_text} is required")
                return 1
            values[config_name] = value

        # Get the target profile name
        target_profile = self._get_target_profile(parsed_args, values['mfa_serial'])

        # Create STS client with the provided credentials
        session = botocore.session.Session()
        sts_client = session.create_client(
            'sts',
            aws_access_key_id=values['aws_access_key_id'],
            aws_secret_access_key=values['aws_secret_access_key'],
        )

        # Call STS to get temporary credentials
        response = self._call_sts_get_session_token(
            sts_client, duration_seconds, values['mfa_serial'], values['mfa_token']
        )
        if not response:
            return 1

        # Write credentials and return
        return self._write_temporary_credentials(
            response['Credentials'], target_profile
        )


# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""
This customization allows the user to specify the values "tcp", "udp",
or "icmp" as values for the --protocol parameter.  The actual Protocol
parameter of the operation accepts only integer protocol numbers.
"""


def _fix_args(params, **kwargs):
    key_name = 'Protocol'
    if key_name in params:
        if params[key_name] == 'tcp':
            params[key_name] = '6'
        elif params[key_name] == 'udp':
            params[key_name] = '17'
        elif params[key_name] == 'icmp':
            params[key_name] = '1'
        elif params[key_name] == 'all':
            params[key_name] = '-1'


def register_protocol_args(cli):
    cli.register('before-parameter-build.ec2.CreateNetworkAclEntry', _fix_args)
    cli.register(
        'before-parameter-build.ec2.ReplaceNetworkAclEntry', _fix_args
    )



# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from datetime import timedelta

from dateutil import parser, tz

DATE_FORMAT = "%Y%m%dT%H%M%SZ"


def get_account_id_from_arn(trail_arn):
    """Gets the account ID portion of an ARN"""
    return trail_arn.split(":")[4]


def get_trail_by_arn(cloudtrail_client, trail_arn):
    """Gets trail information based on the trail's ARN"""
    trails = cloudtrail_client.describe_trails()["trailList"]
    for trail in trails:
        if trail.get("TrailARN", None) == trail_arn:
            return trail
    raise ValueError("A trail could not be found for %s" % trail_arn)


def format_date(date):
    """Returns a formatted date string in a CloudTrail date format"""
    return date.strftime(DATE_FORMAT)


def normalize_date(date):
    """Returns a normalized date using a UTC timezone"""
    return date.replace(tzinfo=tz.tzutc())


def parse_date(date_string):
    try:
        return parser.parse(date_string)
    except ValueError:
        raise ValueError("Unable to parse date value: %s" % date_string)


class PublicKeyProvider:
    """Retrieves public key from CloudTrail with Fingerprint (ID)."""

    def __init__(self, cloudtrail_client):
        self._cloudtrail_client = cloudtrail_client

    def get_public_keys(self, start_date, end_date):
        """Loads public keys in a date range into a returned dict.

        :type start_date: datetime
        :param start_date: Start date of a date range.
        :type end_date: datetime
        :param end_date: End date of a date range.
        :rtype: dict
        :return: Returns a dict where each key is the fingerprint of the
            public key, and each value is a dict of public key data.
        """
        public_keys = self._cloudtrail_client.list_public_keys(
            StartTime=start_date, EndTime=end_date
        )
        public_keys_in_range = public_keys["PublicKeyList"]
        return dict((key["Fingerprint"], key) for key in public_keys_in_range)

    def get_public_key(self, signature_generate_time, public_key_fingerprint):
        """Loads public key that been used in a specific time.

        :type signature_generate_time: datetime
        :param signature_generate_time: Time for the key been used for signing.
        :type public_key_fingerprint: string
        :param public_key_fingerprint: Fingerprint (ID) of a public key.

        :rtype: string
        :return: Return the public key value.
        """
        search_end_date = format_date(
            normalize_date(signature_generate_time + timedelta(days=20))
        )
        public_keys = self._cloudtrail_client.list_public_keys(
            StartTime=signature_generate_time, EndTime=search_end_date
        )
        public_keys_in_range = public_keys["PublicKeyList"]
        for key in public_keys_in_range:
            if key["Fingerprint"] == public_key_fingerprint:
                return key["Value"]

        raise RuntimeError(
            "No public keys found for key with fingerprint: %s"
            % public_key_fingerprint
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

from awscli.autocomplete import LazyClientCreator
from awscli.autocomplete.completer import BaseCompleter, CompletionResult
from awscli.autocomplete.filters import startswith_filter

LOG = logging.getLogger(__name__)


class ServerSideCompleter(BaseCompleter):
    def __init__(
        self, completion_lookup, client_creator, response_filter=None
    ):
        self._completion_lookup = completion_lookup
        self._client_creator = client_creator
        self._jmespath = None
        self._filter = response_filter
        if self._filter is None:
            self._filter = startswith_filter

    @property
    def jmespath(self):
        if self._jmespath is None:
            import jmespath

            self._jmespath = jmespath
        return self._jmespath

    def complete(self, parsed):
        """

        :param parsed: The parsed CLI command the user has typed.
        :type parsed:  awscli.autocomplete.parser.ParsedResult
        :return: A list of completion results, or None.
        :rtype: List[awscli.autocomplete.completer.CompletionResult]

        """
        if parsed.unparsed_items or parsed.current_fragment is None:
            return
        # This completer only applies to option values.  If the current fragment
        # we're on isn't an option value we can short circuit and return.
        if not self._on_cli_option_value_fragment(parsed):
            return
        completion_data = self._completion_lookup.get_server_completion_data(
            parsed.lineage,
            parsed.current_command,
            parsed.current_param,
        )
        if completion_data is None:
            return
        # At this point we know that this completer matches and we have
        # the information we need to actually generate completion suggestions.
        # We also know that we're only generate completoins that have a single
        # list element.  We'll need to update this once we support completions
        # with a 'parameters' key.
        raw_results = self._retrieve_remote_completion_data(
            parsed, completion_data['completions'][0]
        )
        return self._filter(
            parsed.current_fragment,
            self._convert_to_completion_data(raw_results, parsed),
        )

    def _convert_to_completion_data(self, raw_results, parsed):
        index = -len(parsed.current_fragment)
        return [CompletionResult(r, index) for r in raw_results]

    def _on_cli_option_value_fragment(self, parsed):
        if parsed.current_param is None or parsed.current_fragment is None:
            return False
        return True

    def _get_client(self, service_name, parsed):
        return self._client_creator.create_client(
            service_name,
            parsed_region=parsed.global_params.get('region'),
            parsed_profile=parsed.global_params.get('profile'),
        )

    def _retrieve_remote_completion_data(self, parsed, completion_data):
        # TODO: Handle global params that can affect how we create the client
        # (region, endpoint-url, profile, timeouts, etc).
        service_name = completion_data['service']
        client = self._get_client(service_name, parsed)
        method_name = completion_data['operation']
        api_params = self._map_command_to_api_params(parsed, completion_data)
        response = self._invoke_api(client, method_name, api_params)
        if response:
            result = self.jmespath.search(completion_data['jp_expr'], response)
            if result:
                return result
        return []

    def _invoke_api(self, client, py_name, api_params):
        try:
            return getattr(client, py_name)(**api_params)
        except Exception:
            # We don't want tracebacks to propagate back out to the user
            # so the best we can do is log the exception.
            LOG.debug(
                "Exception raised when calling %s on client %s",
                client,
                py_name,
                exc_info=True,
            )
            return {}

    def _map_command_to_api_params(self, parsed, completion_data):
        # Right now we don't autogenerate any completion with the 'parameters'
        # key populated, but we will eventually need to handle this once we
        # do so.
        return {}


class BaseCustomServerSideCompleter(ServerSideCompleter):
    _PARAM_NAME = ''
    _COMMAND_NAMES = ''
    _LINEAGE = []

    def __init__(self, client_creator=None):
        self._client_creator = client_creator
        if self._client_creator is None:
            self._client_creator = LazyClientCreator()

    def complete(self, parsed):
        if not self._is_value_for_param(parsed):
            return
        remote_results = self._get_remote_results(parsed)
        completion_results = []
        for remote_result in remote_results:
            if remote_result.startswith(parsed.current_fragment):
                completion_results.append(
                    CompletionResult(
                        remote_result, -len(parsed.current_fragment)
                    )
                )
        return completion_results

    def _is_value_for_param(self, parsed):
        return (
            parsed.lineage == self._LINEAGE
            and parsed.current_command in self._COMMAND_NAMES
            and parsed.current_param == self._PARAM_NAME
        )

    def _get_remote_results(self, parsed):
        raise NotImplementedError('_get_remote_results()')



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import functools
import json
import logging
import os
import sys

from botocore.client import Config

from awscli.compat import compat_open, get_stdout_text_writer
from awscli.customizations.cloudformation import exceptions
from awscli.customizations.cloudformation.deployer import Deployer
from awscli.customizations.cloudformation.yamlhelper import yaml_parse
from awscli.customizations.commands import BasicCommand
from awscli.customizations.exceptions import ParamValidationError
from awscli.customizations.s3uploader import S3Uploader
from awscli.utils import write_exception

LOG = logging.getLogger(__name__)


class BaseParameterOverrideParser:
    def can_parse(self, data):
        # Returns true/false if it can parse
        raise NotImplementedError('can_parse')

    def parse(self, data):
        # Return the properly formatted parameter dictionary
        raise NotImplementedError('parse')


class CodePipelineLikeParameterOverrideParser(BaseParameterOverrideParser):
    def can_parse(self, data):
        return isinstance(data, dict) and 'Parameters' in data

    def parse(self, data):
        # Parse parameter_overrides if they were given in
        # CodePipeline params file format
        # {
        #     "Parameters": {
        #         "ParameterKey": "ParameterValue"
        #     }
        # }
        return data['Parameters']


class CloudFormationLikeParameterOverrideParser(BaseParameterOverrideParser):
    def can_parse(self, data):
        for param_pair in data:
            if (
                'ParameterKey' not in param_pair
                or 'ParameterValue' not in param_pair
            ):
                return False
            if len(param_pair.keys()) > 2:
                return False
        return True

    def parse(self, data):
        # Parse parameter_overrides if they were given in
        # CloudFormation params file format
        # [{
        #   "ParameterKey": "string",
        #   "ParameterValue": "string",
        # }]
        return {
            param['ParameterKey']: param['ParameterValue'] for param in data
        }


class StringEqualsParameterOverrideParser(BaseParameterOverrideParser):
    def can_parse(self, data):
        return all(
            isinstance(param, str) and len(param.split("=", 1)) == 2
            for param in data
        )

    def parse(self, data):
        result = {}
        for param in data:
            # Split at first '=' from left
            key_value_pair = param.split("=", 1)
            result[key_value_pair[0]] = key_value_pair[1]
        return result


class DeployCommand(BasicCommand):
    MSG_NO_EXECUTE_CHANGESET = (
        "Changeset created successfully. Run the following command to "
        "review changes:"
        "\n"
        "aws cloudformation describe-change-set --change-set-name "
        "{changeset_id}"
        "\n"
    )

    MSG_EXECUTE_SUCCESS = "Successfully created/updated stack - {stack_name}\n"

    PARAMETER_OVERRIDE_CMD = "parameter-overrides"
    TAGS_CMD = "tags"

    NAME = 'deploy'
    DESCRIPTION = BasicCommand.FROM_FILE(
        "cloudformation", "_deploy_description.rst"
    )

    ARG_TABLE = [
        {
            'name': 'template-file',
            'required': True,
            'help_text': (
                'The path where your AWS CloudFormation'
                ' template is located.'
            ),
        },
        {
            'name': 'stack-name',
            'action': 'store',
            'required': True,
            'help_text': (
                'The name of the AWS CloudFormation stack you\'re deploying to.'
                ' If you specify an existing stack, the command updates the'
                ' stack. If you specify a new stack, the command creates it.'
            ),
        },
        {
            'name': 's3-bucket',
            'required': False,
            'help_text': (
                'The name of the S3 bucket where this command uploads your '
                'CloudFormation template. This is required the deployments of '
                'templates sized greater than 51,200 bytes'
            ),
        },
        {
            "name": "force-upload",
            "action": "store_true",
            "help_text": (
                'Indicates whether to override existing files in the S3 bucket.'
                ' Specify this flag to upload artifacts even if they '
                ' match existing artifacts in the S3 bucket.'
            ),
        },
        {
            'name': 's3-prefix',
            'help_text': (
                'A prefix name that the command adds to the'
                ' artifacts\' name when it uploads them to the S3 bucket.'
                ' The prefix name is a path name (folder name) for'
                ' the S3 bucket.'
            ),
        },
        {
            'name': 'kms-key-id',
            'help_text': (
                'The ID of an AWS KMS key that the command uses'
                ' to encrypt artifacts that are at rest in the S3 bucket.'
            ),
        },
        {
            'name': PARAMETER_OVERRIDE_CMD,
            'action': 'store',
            'required': False,
            'nargs': '+',
            'default': [],
            'help_text': (
                'A list of parameter structures that specify input parameters'
                ' for your stack template. If you\'re updating a stack and you'
                ' don\'t specify a parameter, the command uses the stack\'s'
                ' existing value. For new stacks, you must specify'
                ' parameters that don\'t have a default value.'
                ' Syntax: ParameterKey1=ParameterValue1'
                ' ParameterKey2=ParameterValue2 ... or JSON file (see Examples)'
            ),
        },
        {
            'name': 'capabilities',
            'action': 'store',
            'required': False,
            'schema': {
                'type': 'array',
                'items': {
                    'type': 'string',
                    'enum': ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'],
                },
            },
            'default': [],
            'help_text': (
                'A list of capabilities that you must specify before AWS'
                ' Cloudformation can create certain stacks. Some stack'
                ' templates might include resources that can affect'
                ' permissions in your AWS account, for example, by creating'
                ' new AWS Identity and Access Management (IAM) users. For'
                ' those stacks, you must explicitly acknowledge their'
                ' capabilities by specifying this parameter. '
                ' The only valid values are CAPABILITY_IAM and'
                ' CAPABILITY_NAMED_IAM. If you have IAM resources, you can'
                ' specify either capability. If you have IAM resources with'
                ' custom names, you must specify CAPABILITY_NAMED_IAM. If you'
                ' don\'t specify this parameter, this action returns an'
                ' InsufficientCapabilities error.'
            ),
        },
        {
            'name': 'no-execute-changeset',
            'action': 'store_false',
            'dest': 'execute_changeset',
            'required': False,
            'help_text': (
                'Indicates whether to execute the change set. Specify this'
                ' flag if you want to view your stack changes before'
                ' executing the change set. The command creates an'
                ' AWS CloudFormation change set and then exits without'
                ' executing the change set. After you view the change set,'
                ' execute it to implement your changes.'
            ),
        },
        {
            'name': 'disable-rollback',
            'required': False,
            'action': 'store_true',
            'group_name': 'disable-rollback',
            'dest': 'disable_rollback',
            'default': False,
            'help_text': (
                'Preserve the state of previously provisioned resources when '
                'the execute-change-set operation fails.'
            ),
        },
        {
            'name': 'no-disable-rollback',
            'required': False,
            'action': 'store_false',
            'group_name': 'disable-rollback',
            'dest': 'disable_rollback',
            'default': True,
            'help_text': (
                'Roll back all resource changes when the execute-change-set '
                'operation fails.'
            ),
        },
        {
            'name': 'role-arn',
            'required': False,
            'help_text': (
                'The Amazon Resource Name (ARN) of an AWS Identity and Access '
                'Management (IAM) role that AWS CloudFormation assumes when '
                'executing the change set.'
            ),
        },
        {
            'name': 'notification-arns',
            'required': False,
            'schema': {'type': 'array', 'items': {'type': 'string'}},
            'help_text': (
                'Amazon Simple Notification Service topic Amazon Resource Names'
                ' (ARNs) that AWS CloudFormation associates with the stack.'
            ),
        },
        {
            'name': 'fail-on-empty-changeset',
            'required': False,
            'action': 'store_true',
            'group_name': 'fail-on-empty-changeset',
            'dest': 'fail_on_empty_changeset',
            'default': False,
            'help_text': (
                'Specify if the CLI should return a non-zero exit code '
                'when there are no changes to be made to the stack. By '
                'default, a zero exit code is returned, and this is '
                'the same behavior that occurs when '
                '`--no-fail-on-empty-changeset` is specified. If '
                '`--fail-on-empty-changeset` is specified, then the '
                'CLI will return a non-zero exit code.'
            ),
        },
        {
            'name': 'no-fail-on-empty-changeset',
            'required': False,
            'action': 'store_false',
            'group_name': 'fail-on-empty-changeset',
            'dest': 'fail_on_empty_changeset',
            'default': False,
            'help_text': (
                'Causes the CLI to return an exit code of 0 if there are no '
                'changes to be made to the stack.'
            ),
        },
        {
            'name': TAGS_CMD,
            'action': 'store',
            'required': False,
            'schema': {'type': 'array', 'items': {'type': 'string'}},
            'default': [],
            'help_text': (
                'A list of tags to associate with the stack that is created'
                ' or updated. AWS CloudFormation also propagates these tags'
                ' to resources in the stack if the resource supports it.'
                ' Syntax: TagKey1=TagValue1 TagKey2=TagValue2 ...'
            ),
        },
    ]

    def _run_main(self, parsed_args, parsed_globals):
        cloudformation_client = self._session.create_client(
            'cloudformation',
            region_name=parsed_globals.region,
            endpoint_url=parsed_globals.endpoint_url,
            verify=parsed_globals.verify_ssl,
        )

        template_dict, template_str, template_size = self.load_template_file(
            parsed_args.template_file
        )

        stack_name = parsed_args.stack_name
        parameter_overrides = self.parse_parameter_overrides(
            parsed_args.parameter_overrides
        )
        tags_dict = self.parse_key_value_arg(parsed_args.tags, self.TAGS_CMD)
        tags = [
            {"Key": key, "Value": value} for key, value in tags_dict.items()
        ]

        parameters = self.merge_parameters(template_dict, parameter_overrides)

        if template_size > 51200 and not parsed_args.s3_bucket:
            raise exceptions.DeployBucketRequiredError()

        bucket = parsed_args.s3_bucket
        if bucket:
            s3_client = self._session.create_client(
                "s3",
                config=Config(signature_version='s3v4'),
                region_name=parsed_globals.region,
                verify=parsed_globals.verify_ssl,
            )

            s3_uploader = S3Uploader(
                s3_client,
                bucket,
                parsed_args.s3_prefix,
                parsed_args.kms_key_id,
                parsed_args.force_upload,
            )
        else:
            s3_uploader = None

        deployer = Deployer(cloudformation_client)
        return self.deploy(
            deployer,
            stack_name,
            template_str,
            parameters,
            parsed_args.capabilities,
            parsed_args.execute_changeset,
            parsed_args.role_arn,
            parsed_args.notification_arns,
            s3_uploader,
            tags,
            parsed_args.fail_on_empty_changeset,
            parsed_args.disable_rollback,
        )

    def load_template_file(self, template_file):
        template_path = os.path.expanduser(template_file)
        if not os.path.isfile(template_path):
            raise exceptions.InvalidTemplatePathError(
                template_path=template_path
            )
        with compat_open(template_path, "r") as handle:
            template_str = handle.read()
        template_dict = yaml_parse(template_str)
        template_size = os.path.getsize(template_path)
        return template_dict, template_str, template_size

    def deploy(
        self,
        deployer,
        stack_name,
        template_str,
        parameters,
        capabilities,
        execute_changeset,
        role_arn,
        notification_arns,
        s3_uploader,
        tags,
        fail_on_empty_changeset=False,
        disable_rollback=False,
    ):
        try:
            result = deployer.create_and_wait_for_changeset(
                stack_name=stack_name,
                cfn_template=template_str,
                parameter_values=parameters,
                capabilities=capabilities,
                role_arn=role_arn,
                notification_arns=notification_arns,
                s3_uploader=s3_uploader,
                tags=tags,
            )
        except exceptions.ChangeEmptyError as ex:
            if fail_on_empty_changeset:
                raise
            write_exception(ex, outfile=get_stdout_text_writer())
            return 0

        if execute_changeset:
            deployer.execute_changeset(
                result.changeset_id, stack_name, disable_rollback
            )
            deployer.wait_for_execute(stack_name, result.changeset_type)
            sys.stdout.write(
                self.MSG_EXECUTE_SUCCESS.format(stack_name=stack_name)
            )
        else:
            sys.stdout.write(
                self.MSG_NO_EXECUTE_CHANGESET.format(
                    changeset_id=result.changeset_id
                )
            )

        sys.stdout.flush()
        return 0

    def merge_parameters(self, template_dict, parameter_overrides):
        """
        CloudFormation CreateChangeset requires a value for every parameter
        from the template, either specifying a new value or use previous value.
        For convenience, this method will accept new parameter values and
        generates a dict of all parameters in a format that ChangeSet API
        will accept

        :param parameter_overrides:
        :return:
        """
        parameter_values = []

        if not isinstance(template_dict.get("Parameters", None), dict):
            return parameter_values

        for key, value in template_dict["Parameters"].items():
            obj = {"ParameterKey": key}

            if key in parameter_overrides:
                obj["ParameterValue"] = parameter_overrides[key]
            else:
                obj["UsePreviousValue"] = True

            parameter_values.append(obj)

        return parameter_values

    def _parse_input_as_json(self, arg_value):
        # In case of reading from file it'll be string and in case
        # of inline json input it'll be list where json string
        # will be the first element
        if arg_value:
            if isinstance(arg_value, str):
                return json.loads(arg_value)
            try:
                return json.loads(arg_value[0])
            except json.JSONDecodeError:
                return None

    def parse_parameter_overrides(self, arg_value):
        data = self._parse_input_as_json(arg_value)
        if data is not None:
            parsers = [
                CloudFormationLikeParameterOverrideParser(),
                CodePipelineLikeParameterOverrideParser(),
                StringEqualsParameterOverrideParser(),
            ]
            for parser in parsers:
                if parser.can_parse(data):
                    return parser.parse(data)
            raise ParamValidationError(
                'JSON passed to --parameter-overrides must be one of '
                'the formats: ["Key1=Value1","Key2=Value2", ...] , '
                '[{"ParameterKey": "Key1", "ParameterValue": "Value1"}, ...] , '
                '["Parameters": {"Key1": "Value1", "Key2": "Value2", ...}]'
            )
        else:
            # In case it was in deploy command format
            # and was input via command line
            return self.parse_key_value_arg(
                arg_value, self.PARAMETER_OVERRIDE_CMD
            )

    def parse_key_value_arg(self, arg_value, argname):
        """
        Converts arguments that are passed as list of "Key=Value" strings
        into a real dictionary.

        :param arg_value list: Array of strings, where each string is of
            form Key=Value
        :param argname string: Name of the argument that contains the value
        :return dict: Dictionary representing the key/value pairs
        """
        result = {}
        for data in arg_value:
            # Split at first '=' from left
            key_value_pair = data.split("=", 1)

            if len(key_value_pair) != 2:
                raise exceptions.InvalidKeyValuePairArgumentError(
                    argname=argname, value=key_value_pair
                )

            result[key_value_pair[0]] = key_value_pair[1]

        return result



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from collections.abc import Mapping, MutableSequence

from awscli.customizations.dynamodb.types import (
    TypeDeserializer,
    TypeSerializer,
)


class ParameterTransformer:
    """Transforms the input to and output from botocore based on shape"""

    def transform(self, params, model, transformation, target_shape):
        """Transforms the dynamodb input to or output from botocore
        It applies a specified transformation whenever a specific shape name
        is encountered while traversing the parameters in the dictionary.
        :param params: The parameters structure to transform.
        :param model: The operation model.
        :param transformation: The function to apply the parameter
        :param target_shape: The name of the shape to apply the
            transformation to
        """
        self._transform_parameters(model, params, transformation, target_shape)

    def _transform_parameters(
        self, model, params, transformation, target_shape
    ):
        type_name = model.type_name
        if type_name in ['structure', 'map', 'list']:
            getattr(self, '_transform_%s' % type_name)(
                model, params, transformation, target_shape
            )

    def _transform_structure(
        self, model, params, transformation, target_shape
    ):
        if not isinstance(params, Mapping):
            return
        for param in params:
            if param in model.members:
                member_model = model.members[param]
                member_shape = member_model.name
                if member_shape == target_shape:
                    params[param] = transformation(params[param])
                else:
                    self._transform_parameters(
                        member_model,
                        params[param],
                        transformation,
                        target_shape,
                    )

    def _transform_map(self, model, params, transformation, target_shape):
        if not isinstance(params, Mapping):
            return
        value_model = model.value
        value_shape = value_model.name
        for key, value in params.items():
            if value_shape == target_shape:
                params[key] = transformation(value)
            else:
                self._transform_parameters(
                    value_model, params[key], transformation, target_shape
                )

    def _transform_list(self, model, params, transformation, target_shape):
        if not isinstance(params, MutableSequence):
            return
        member_model = model.member
        member_shape = member_model.name
        for i, item in enumerate(params):
            if member_shape == target_shape:
                params[i] = transformation(item)
            else:
                self._transform_parameters(
                    member_model, params[i], transformation, target_shape
                )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import sys


def lazy_call(import_name, **kwargs):
    """Import a callable and invoke it with the provided kwargs.

    :param import_name: A dotted string of the form ``package.module.Callable``.
    :type import_name: string
    :param kwargs: kwargs to pass to the callable once it's imported.
    :type kwargs:  dict

    """
    module_name, callable_name = import_name.rsplit('.', 1)
    __import__(module_name)
    module = sys.modules[module_name]
    return getattr(module, callable_name)(**kwargs)


class LazyClientCreator:
    """Lazy create a botocore client.

    This class will defer creating a client until the create_client method
    is invoked.  It will also avoid importing botocore until we need to
    create a client.

    Importing botocore and creating a botocore session/client is an expensive
    process, and we only want to do this when we know for sure that we need
    a client.  This class manages this process.

    """

    def __init__(self, import_name='awscli.clidriver.create_clidriver'):
        self._import_name = import_name
        self._session_cache = {}

    def create_client(
        self, service_name, parsed_region=None, parsed_profile=None, **kwargs
    ):
        if self._session_cache.get(parsed_profile) is None:
            session = self.create_session()
            session.set_config_variable('profile', parsed_profile)
            self._session_cache[parsed_profile] = session
        self._session_cache[parsed_profile].set_config_variable(
            'region', parsed_region
        )
        return self._session_cache[parsed_profile].create_client(
            service_name, **kwargs
        )

    def create_session(self):
        return lazy_call(self._import_name).session



# Copyright 2013-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
# If the user does not specify any overrides,
# these are the default values we use for the s3 transfer
# commands.
import logging

from botocore.utils import ensure_boolean
from s3transfer.manager import TransferConfig

from awscli.customizations.s3 import constants
from awscli.customizations.s3.utils import human_readable_to_int

LOGGER = logging.getLogger(__name__)

DEFAULTS = {
    'multipart_threshold': 8 * (1024**2),
    'multipart_chunksize': 8 * (1024**2),
    'max_concurrent_requests': 10,
    'max_queue_size': 1000,
    'max_bandwidth': None,
    'preferred_transfer_client': constants.AUTO_RESOLVE_TRANSFER_CLIENT,
    'target_bandwidth': None,
    'io_chunksize': 256 * 1024,
    'should_stream': None,
    'disk_throughput': None,
    'direct_io': None,
}


class InvalidConfigError(Exception):
    pass


class RuntimeConfig:
    POSITIVE_INTEGERS = [
        'multipart_chunksize',
        'multipart_threshold',
        'max_concurrent_requests',
        'max_queue_size',
        'max_bandwidth',
        'target_bandwidth',
        'io_chunksize',
        'disk_throughput',
    ]
    HUMAN_READABLE_SIZES = [
        'multipart_chunksize',
        'multipart_threshold',
        'io_chunksize',
    ]
    HUMAN_READABLE_RATES = [
        'max_bandwidth',
        'target_bandwidth',
        'disk_throughput',
    ]
    SUPPORTED_CHOICES = {
        'preferred_transfer_client': [
            constants.AUTO_RESOLVE_TRANSFER_CLIENT,
            constants.CLASSIC_TRANSFER_CLIENT,
            constants.CRT_TRANSFER_CLIENT,
        ]
    }
    CHOICE_ALIASES = {
        'preferred_transfer_client': {
            'default': constants.CLASSIC_TRANSFER_CLIENT
        }
    }
    BOOLEANS = ['should_stream', 'direct_io']

    @staticmethod
    def defaults():
        return DEFAULTS.copy()

    def build_config(self, **kwargs):
        """Create and convert a runtime config dictionary.

        This method will merge and convert S3 runtime configuration
        data into a single dictionary that can then be passed to classes
        that use this runtime config.

        :param kwargs:  Any key in the ``DEFAULTS`` dict.
        :return: A dictionary of the merged and converted values.

        """
        runtime_config = DEFAULTS.copy()
        if kwargs:
            runtime_config.update(kwargs)
        self._convert_human_readable_sizes(runtime_config)
        self._convert_human_readable_rates(runtime_config)
        self._convert_booleans(runtime_config)
        self._resolve_choice_aliases(runtime_config)
        self._validate_config(runtime_config)
        return runtime_config

    def _convert_human_readable_sizes(self, runtime_config):
        for attr in self.HUMAN_READABLE_SIZES:
            value = runtime_config.get(attr)
            if value is not None and not isinstance(value, int):
                runtime_config[attr] = human_readable_to_int(value)

    def _convert_human_readable_rates(self, runtime_config):
        for attr in self.HUMAN_READABLE_RATES:
            value = runtime_config.get(attr)
            if value is not None and not isinstance(value, int):
                if value.endswith('B/s'):
                    runtime_config[attr] = self._human_readable_rate_to_int(
                        value
                    )
                elif value.endswith('b/s'):
                    bits_per_sec = self._human_readable_rate_to_int(value)
                    bytes_per_sec = int(bits_per_sec / 8)
                    runtime_config[attr] = bytes_per_sec
                elif self._is_integer_str(value):
                    runtime_config[attr] = int(value)
                else:
                    raise InvalidConfigError(
                        'Invalid rate: %s. The value must be expressed '
                        'as an integer in terms of bytes per second '
                        '(e.g. 10485760) or a rate in terms of bytes '
                        'per second (e.g. 10MB/s or 800KB/s) or bits per '
                        'second (e.g. 10Mb/s or 800Kb/s)' % value
                    )

    def _convert_booleans(self, runtime_config):
        for attr in self.BOOLEANS:
            value = runtime_config.get(attr)
            if value is not None:
                runtime_config[attr] = ensure_boolean(value)

    def _human_readable_rate_to_int(self, value):
        # The human_readable_to_int() utility only supports integers (e.g. 1024)
        # as strings and human readable sizes (e.g. 10MB, 5GB). It does not
        # directly support human readable rates (e.g. 10MB/s, 5GB/s) nor human
        # readable sizes that do not contain a magnitude prefix (e.g. 1024B).
        # However, the rate configuration require the values end with "/s"
        # and allows for values that do not have a magnitude prefix
        # (e.g. 1024B/s).
        #
        # To account for these limitations:
        #
        # 1. If the human readable rate does not contain a magnitude prefix, it
        #    will strip the "B/s" to provide the value as an integer string to
        #    human_readable_int() (e.g. "1024B/s" -> "1024")
        #
        # 2. Otherwise, it will strip the "/s" to provide the value as a
        #    human readable size to human_readable_int()
        #    (e.g. "1024MB/s -> "1024MB")
        if self._is_integer_str(value[:-3]):
            return human_readable_to_int(value[:-3])
        return human_readable_to_int(value[:-2])

    def _is_integer_str(self, value):
        try:
            int(value)
            return True
        except ValueError:
            return False

    def _resolve_choice_aliases(self, runtime_config):
        for attr in self.CHOICE_ALIASES:
            current_value = runtime_config.get(attr)
            if current_value in self.CHOICE_ALIASES[attr]:
                resolved_value = self.CHOICE_ALIASES[attr][current_value]
                LOGGER.debug(
                    'Resolved %s configuration alias value "%s" to "%s"',
                    attr,
                    current_value,
                    resolved_value,
                )
                runtime_config[attr] = resolved_value

    def _validate_config(self, runtime_config):
        self._validate_positive_integers(runtime_config)
        self._validate_choices(runtime_config)

    def _validate_positive_integers(self, runtime_config):
        for attr in self.POSITIVE_INTEGERS:
            value = runtime_config.get(attr)
            if value is not None:
                try:
                    runtime_config[attr] = int(value)
                    if not runtime_config[attr] > 0:
                        self._error_positive_value(attr, value)
                except ValueError:
                    self._error_positive_value(attr, value)

    def _validate_choices(self, runtime_config):
        for attr in self.SUPPORTED_CHOICES:
            value = runtime_config.get(attr)
            if value is not None:
                if value not in self.SUPPORTED_CHOICES[attr]:
                    self._error_invalid_choice(attr, value)

    def _error_positive_value(self, name, value):
        raise InvalidConfigError(
            "Value for %s must be a positive integer: %s" % (name, value)
        )

    def _error_invalid_choice(self, name, value):
        raise InvalidConfigError(
            f'Invalid value: "{value}" for configuration option: "{name}". '
            f'Supported values are: {", ".join(self.SUPPORTED_CHOICES[name])}'
        )


def create_transfer_config_from_runtime_config(runtime_config):
    """
    Creates an equivalent s3transfer TransferConfig

    :type runtime_config: dict
    :argument runtime_config: A valid RuntimeConfig-generated dict.

    :returns: A TransferConfig with the same configuration as the runtime
        config.
    """
    translation_map = {
        'max_concurrent_requests': 'max_request_concurrency',
        'max_queue_size': 'max_request_queue_size',
        'multipart_threshold': 'multipart_threshold',
        'multipart_chunksize': 'multipart_chunksize',
        'max_bandwidth': 'max_bandwidth',
        'io_chunksize': 'io_chunksize',
    }
    kwargs = {}
    for key, value in runtime_config.items():
        if key not in translation_map:
            continue
        kwargs[translation_map[key]] = value
    return TransferConfig(**kwargs)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.autocomplete.db import DatabaseConnection


def create_model_indexer(filename):
    index = ModelIndexer(DatabaseConnection(filename))
    return index


class ModelIndexer:
    # TODO add full names to custom commands to get rid of this map
    _HIGH_LEVEL_SERVICE_FULL_NAMES = {
        's3': 'High level S3 commands',
        'ddb': 'High level DynamoDB commands',
    }

    _NON_SERVICE_COMMANDS = ['configure', 'history', 'cli-dev', 'login', 'logout']

    _CREATE_CMD_TABLE = """\
        CREATE TABLE IF NOT EXISTS command_table (
          command TEXT,
          full_name TEXT,
          parent TEXT REFERENCES command_table,
          PRIMARY KEY (command, parent)
        );
    """
    _CREATE_PARAM_TABLE = """\
        CREATE TABLE IF NOT EXISTS param_table (
          param_id INTEGER PRIMARY KEY,
          argname TEXT,
          type_name TEXT,
          command TEXT,
          parent TEXT,
          nargs TEXT,
          positional_arg TEXT,
          required INTEGER,
          FOREIGN KEY (command, parent) REFERENCES
            command_table(command, parent)
        );
    """

    _CREATE_COMMAND_TABLE_INDEX = """\
        CREATE INDEX parent_index
            ON command_table(parent);
    """

    _CREATE_PARAM_TABLE_INDEX = """\
        CREATE INDEX parent_command_index
            ON param_table(parent, command);
    """

    def __init__(self, db_connection):
        self._db_connection = db_connection

    def generate_index(self, clidriver):
        self._create_tables()
        parent = 'aws'
        self._db_connection.execute(
            'INSERT OR REPLACE INTO command_table (command, parent)'
            'VALUES (:command, :parent)',
            command=parent,
            parent='',
        )
        help_command_table = clidriver.create_help_command().command_table
        command_table = clidriver.subcommand_table
        self._generate_arg_index(
            command=parent, parent='', arg_table=clidriver.arg_table
        )
        self._generate_command_index(
            command_table, parent=parent, help_command_table=help_command_table
        )

        self._generate_table_indexes()

    def _create_tables(self):
        self._db_connection.execute(self._CREATE_CMD_TABLE)
        self._db_connection.execute(self._CREATE_PARAM_TABLE)

    def _generate_arg_index(self, command, parent, arg_table):
        for name, value in arg_table.items():
            # SQLite has no boolean data type, so we use Integer instead
            required = 1 if value.required else 0
            self._db_connection.execute(
                'INSERT INTO param_table '
                '(argname, type_name, command, parent, nargs, positional_arg,'
                'required)'
                ' VALUES (:argname, :type_name, :command, :parent, :nargs, '
                '         :positional_arg, :required)',
                argname=name,
                type_name=value.cli_type_name,
                command=command,
                parent=parent,
                nargs=value.nargs,
                positional_arg=value.positional_arg,
                required=required,
            )

    def _get_service_full_name(self, name, help_command_table):
        if help_command_table and name not in self._NON_SERVICE_COMMANDS:
            if name in self._HIGH_LEVEL_SERVICE_FULL_NAMES:
                return self._HIGH_LEVEL_SERVICE_FULL_NAMES[name]
            service = help_command_table.get(name)
            if service:
                return service.service_model.metadata['serviceFullName']

    def _generate_command_index(
        self, command_table, parent, help_command_table=None
    ):
        for name, command in command_table.items():
            full_name = self._get_service_full_name(name, help_command_table)
            self._db_connection.execute(
                'INSERT INTO command_table (command, parent, full_name) '
                'VALUES (:command, :parent, :full_name)',
                command=name,
                parent=parent,
                full_name=full_name,
            )
            self._generate_arg_index(
                command=name, parent=parent, arg_table=command.arg_table
            )
            self._generate_command_index(
                command.subcommand_table, parent=f'{parent}.{name}'
            )

    def _generate_table_indexes(self):
        self._db_connection.execute(self._CREATE_COMMAND_TABLE_INDEX)
        self._db_connection.execute(self._CREATE_PARAM_TABLE_INDEX)



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


def register_create_hosted_zone_doc_fix(cli):
    # We can remove this customization once we begin documenting
    # members of complex parameters because the member's docstring
    # has the necessary documentation.
    cli.register(
        'doc-option.route53.create-hosted-zone.hosted-zone-config',
        add_private_zone_note,
    )


def add_private_zone_note(help_command, **kwargs):
    note = (
        '<p>Note do <b>not</b> include <code>PrivateZone</code> in this '
        'input structure. Its value is returned in the output to the command.'
        '</p>'
    )
    help_command.doc.include_doc_string(note)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

from awscli.customizations.s3.subcommands import NO_OVERWRITE
from awscli.customizations.s3.syncstrategy.base import BaseSync

LOG = logging.getLogger(__name__)


class NoOverwriteSync(BaseSync):
    """Sync strategy that prevents overwriting of existing files at the destination.
    This strategy is used only for files that exist at both source and destination
    (file_at_src_and_dest_sync_strategy). It always returns False to prevent any
    overwriting of existing files, regardless of size or modification time differences.
    """

    ARGUMENT = NO_OVERWRITE

    def determine_should_sync(self, src_file, dest_file):
        LOG.debug(
            f"warning: skipping {src_file.src} -> {src_file.dest}, file exists at destination"
        )
        return False



# Copyright 2012-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import sys

from botocore.compat import json

from awscli.customizations.servicecatalog import helptext
from awscli.customizations.servicecatalog.generatebase import (
    GenerateBaseCommand,
)


class GenerateProductCommand(GenerateBaseCommand):
    NAME = "product"
    DESCRIPTION = helptext.PRODUCT_COMMAND_DESCRIPTION
    ARG_TABLE = [
        {
            'name': 'product-name',
            'required': True,
            'help_text': helptext.PRODUCT_NAME,
        },
        {
            'name': 'product-owner',
            'required': True,
            'help_text': helptext.OWNER,
        },
        {
            'name': 'product-type',
            'required': True,
            'help_text': helptext.PRODUCT_TYPE,
            'choices': ['CLOUD_FORMATION_TEMPLATE', 'MARKETPLACE'],
        },
        {
            'name': 'product-description',
            'required': False,
            'help_text': helptext.PRODUCT_DESCRIPTION,
        },
        {
            'name': 'product-distributor',
            'required': False,
            'help_text': helptext.DISTRIBUTOR,
        },
        {
            'name': 'tags',
            'required': False,
            'schema': {'type': 'array', 'items': {'type': 'string'}},
            'default': [],
            'synopsis': '--tags Key=key1,Value=value1 Key=key2,Value=value2',
            'help_text': helptext.TAGS,
        },
        {
            'name': 'file-path',
            'required': True,
            'help_text': helptext.FILE_PATH,
        },
        {
            'name': 'bucket-name',
            'required': True,
            'help_text': helptext.BUCKET_NAME,
        },
        {
            'name': 'support-description',
            'required': False,
            'help_text': helptext.SUPPORT_DESCRIPTION,
        },
        {
            'name': 'support-email',
            'required': False,
            'help_text': helptext.SUPPORT_EMAIL,
        },
        {
            'name': 'provisioning-artifact-name',
            'required': True,
            'help_text': helptext.PA_NAME,
        },
        {
            'name': 'provisioning-artifact-description',
            'required': True,
            'help_text': helptext.PA_DESCRIPTION,
        },
        {
            'name': 'provisioning-artifact-type',
            'required': True,
            'help_text': helptext.PA_TYPE,
            'choices': [
                'CLOUD_FORMATION_TEMPLATE',
                'MARKETPLACE_AMI',
                'MARKETPLACE_CAR',
            ],
        },
    ]

    def _run_main(self, parsed_args, parsed_globals):
        super(GenerateProductCommand, self)._run_main(
            parsed_args, parsed_globals
        )
        self.region = self.get_and_validate_region(parsed_globals)

        self.s3_url = self.create_s3_url(
            parsed_args.bucket_name, parsed_args.file_path
        )
        self.scs_client = self._session.create_client(
            'servicecatalog',
            region_name=self.region,
            endpoint_url=parsed_globals.endpoint_url,
            verify=parsed_globals.verify_ssl,
        )

        response = self.create_product(
            self.build_args(parsed_args, self.s3_url), parsed_globals
        )
        sys.stdout.write(json.dumps(response, indent=2, ensure_ascii=False))

        return 0

    def create_product(self, args, parsed_globals):
        response = self.scs_client.create_product(**args)
        if 'ResponseMetadata' in response:
            del response['ResponseMetadata']
        return response

    def _extract_tags(self, args_tags):
        tags = []
        for tag in args_tags:
            tags.append(dict(t.split('=') for t in tag.split(',')))
        return tags

    def build_args(self, parsed_args, s3_url):
        args = {
            "Name": parsed_args.product_name,
            "Owner": parsed_args.product_owner,
            "ProductType": parsed_args.product_type,
            "Tags": self._extract_tags(parsed_args.tags),
            "ProvisioningArtifactParameters": {
                'Name': parsed_args.provisioning_artifact_name,
                'Description': parsed_args.provisioning_artifact_description,
                'Info': {'LoadTemplateFromURL': s3_url},
                'Type': parsed_args.provisioning_artifact_type,
            },
        }

        # Non-required args
        if parsed_args.support_description:
            args["SupportDescription"] = parsed_args.support_description
        if parsed_args.product_description:
            args["Description"] = parsed_args.product_description
        if parsed_args.support_email:
            args["SupportEmail"] = parsed_args.support_email
        if parsed_args.product_distributor:
            args["Distributor"] = parsed_args.product_distributor

        return args



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from awscli.customizations.login.login import LoginCommand
from awscli.customizations.login.logout import LogoutCommand


def register_login_cmds(cli):
    cli.register('building-command-table.main', LoginCommand.add_command)
    cli.register('building-command-table.main', LogoutCommand.add_command)



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations import utils
from awscli.customizations.codedeploy.deregister import Deregister
from awscli.customizations.codedeploy.install import Install
from awscli.customizations.codedeploy.locationargs import (
    modify_revision_arguments,
)
from awscli.customizations.codedeploy.push import Push
from awscli.customizations.codedeploy.register import Register
from awscli.customizations.codedeploy.uninstall import Uninstall


def initialize(cli):
    """
    The entry point for CodeDeploy high level commands.
    """
    cli.register('building-command-table.main', change_name)
    cli.register('building-command-table.deploy', inject_commands)
    cli.register(
        'building-argument-table.deploy.get-application-revision',
        modify_revision_arguments,
    )
    cli.register(
        'building-argument-table.deploy.register-application-revision',
        modify_revision_arguments,
    )
    cli.register(
        'building-argument-table.deploy.create-deployment',
        modify_revision_arguments,
    )


def change_name(command_table, session, **kwargs):
    """
    Change all existing 'aws codedeploy' commands to 'aws deploy' commands.
    """
    utils.rename_command(command_table, 'codedeploy', 'deploy')


def inject_commands(command_table, session, **kwargs):
    """
    Inject custom 'aws deploy' commands.
    """
    command_table['push'] = Push(session)
    command_table['register'] = Register(session)
    command_table['deregister'] = Deregister(session)
    command_table['install'] = Install(session)
    command_table['uninstall'] = Uninstall(session)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import re

from awscli.customizations.commands import BasicCommand
from awscli.customizations.configure.writer import ConfigFileWriter
from awscli.customizations.exceptions import ParamValidationError
from awscli.customizations.utils import validate_mutually_exclusive

from . import (
    PREDEFINED_SECTION_NAMES,
    SUBSECTION_TYPE_ALLOWLIST,
    get_section_header,
    profile_to_section,
)


class ConfigureSetCommand(BasicCommand):
    NAME = 'set'
    DESCRIPTION = BasicCommand.FROM_FILE(
        'configure', 'set', '_description.rst'
    )
    SYNOPSIS = 'aws configure set varname value [--profile profile-name]'
    EXAMPLES = BasicCommand.FROM_FILE('configure', 'set', '_examples.rst')
    ARG_TABLE = [
        {
            'name': 'varname',
            'help_text': 'The name of the config value to set.',
            'action': 'store',
            'cli_type_name': 'string',
            'positional_arg': True,
        },
        {
            'name': 'value',
            'help_text': 'The value to set.',
            'action': 'store',
            'no_paramfile': True,  # To disable the default paramfile behavior
            'cli_type_name': 'string',
            'positional_arg': True,
        },
        {
            'name': 'sso-session',
            'help_text': 'The name of the sub-section to configure.',
            'action': 'store',
            'cli_type_name': 'string',
            'group_name': 'subsection',
        },
        {
            'name': 'services',
            'help_text': 'The name of the sub-section to configure.',
            'action': 'store',
            'cli_type_name': 'string',
            'group_name': 'subsection',
        },
    ]
    # Any variables specified in this list will be written to
    # the ~/.aws/credentials file instead of ~/.aws/config.
    _WRITE_TO_CREDS_FILE = [
        'aws_access_key_id',
        'aws_secret_access_key',
        'aws_session_token',
    ]

    def __init__(self, session, config_writer=None):
        super().__init__(session)
        if config_writer is None:
            config_writer = ConfigFileWriter()
        self._config_writer = config_writer

    def _get_config_file(self, path):
        config_path = self._session.get_config_variable(path)
        return os.path.expanduser(config_path)

    def _subsection_parameter_to_argument_name(self, parameter_name):
        return parameter_name.replace("-", "_")

    def _get_subsection_from_args(self, args):
        # Validate mutual exclusivity of sub-section type parameters
        groups = [
            [self._subsection_parameter_to_argument_name(key)]
            for key in SUBSECTION_TYPE_ALLOWLIST.keys()
        ]
        validate_mutually_exclusive(args, *groups)

        subsection_name = None
        subsection_type = None

        for section_type in SUBSECTION_TYPE_ALLOWLIST.keys():
            cli_parameter_name = self._subsection_parameter_to_argument_name(
                section_type
            )
            if hasattr(args, cli_parameter_name):
                subsection_name = getattr(args, cli_parameter_name)
                if subsection_name is not None:
                    if not re.match(r"[\w\d_\-/.%@:\+]+", subsection_name):
                        raise ParamValidationError(
                            f"aws: [ERROR]: Invalid value for --{section_type}."
                        )
                    subsection_type = section_type
                    break

        return (subsection_type, subsection_name)

    def _set_subsection_property(
        self, section_type, section_name, varname, value
    ):
        if '.' in varname:
            parts = varname.split('.')
            # Check if there are more than two parts to the property name to set (e.g., aaa.bbb.ccc)
            # This would result in a deeply nested property, which is not supported.
            if len(parts) > 2:
                raise ParamValidationError(
                    "Found more than two parts in the property to set. "
                    "Deep nesting of properties is not supported."
                )
            varname = parts[0]
            value = {parts[1]: value}

        # Build update dict
        updated_config = {
            '__section__': get_section_header(section_type, section_name),
            varname: value,
        }

        # Write to config file
        config_filename = self._get_config_file('config_file')
        self._config_writer.update_config(updated_config, config_filename)

        return 0

    def _run_main(self, args, parsed_globals):
        varname = args.varname
        value = args.value
        profile = 'default'

        section_type, section_name = self._get_subsection_from_args(args)
        if section_type is not None:
            return self._set_subsection_property(
                section_type, section_name, varname, value
            )

        # Not in a sub-section, continue with previous profile logic.
        # Before handing things off to the config writer,
        # we need to find out three things:
        # 1. What section we're writing to (profile).
        # 2. The name of the config key (varname)
        # 3. The actual value (value).
        if '.' not in varname:
            # unqualified name, scope it to the current
            # profile (or leave it as the 'default' section if
            # no profile is set).
            if self._session.profile is not None:
                profile = self._session.profile
        else:
            # First figure out if it's been scoped to a profile.
            parts = varname.split('.')
            if parts[0] in ('default', 'profile'):
                # Then we know we're scoped to a profile.
                if parts[0] == 'default':
                    profile = 'default'
                    remaining = parts[1:]
                else:
                    # [profile, profile_name, ...]
                    profile = parts[1]
                    remaining = parts[2:]
                varname = remaining[0]
                if len(remaining) == 2:
                    value = {remaining[1]: value}
            elif parts[0] not in PREDEFINED_SECTION_NAMES:
                if self._session.profile is not None:
                    profile = self._session.profile
                else:
                    profile_name = self._session.get_config_variable('profile')
                    if profile_name is not None:
                        profile = profile_name
                varname = parts[0]
                if len(parts) == 2:
                    value = {parts[1]: value}
            elif len(parts) == 2:
                # Otherwise it's something in the [plugin] section
                profile, varname = parts
        config_filename = self._get_config_file('config_file')
        if varname in self._WRITE_TO_CREDS_FILE:
            # When writing to the creds file, the section is just the profile
            section = profile
            config_filename = self._get_config_file('credentials_file')
        elif profile in PREDEFINED_SECTION_NAMES or profile == 'default':
            section = profile
        else:
            section = profile_to_section(profile)
        updated_config = {'__section__': section, varname: value}
        self._config_writer.update_config(updated_config, config_filename)
        return 0



# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

# A list of auth types supported by the signers in botocore/crt/auth.py. This
# should always match the keys of botocore.crt.auth.CRT_AUTH_TYPE_MAPS. The
# information is duplicated here so that it can be accessed in environments
# where `awscrt` is not present and any import from botocore.crt.auth would
# fail.



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import re

import jmespath

from awscli.arguments import CustomArgument
from awscli.compat import compat_open
from awscli.customizations.exceptions import ParamValidationError


def resolve_given_outfile_path(path):
    """Asserts that a path is writable and returns the expanded path"""
    if path is None:
        return
    outfile = os.path.expanduser(os.path.expandvars(path))
    if not os.access(os.path.dirname(os.path.abspath(outfile)), os.W_OK):
        raise ParamValidationError('Unable to write to file: %s' % outfile)
    return outfile


def is_parsed_result_successful(parsed_result):
    """Returns True if a parsed result is successful"""
    return parsed_result['ResponseMetadata']['HTTPStatusCode'] < 300


class OverrideRequiredArgsArgument(CustomArgument):
    """An argument that if specified makes all other arguments not required

    By not required, it refers to not having an error thrown when the
    parser does not find an argument that is required on the command line.
    To obtain this argument's property of ignoring required arguments,
    subclass from this class and fill out the ``ARG_DATA`` parameter as
    described below. Note this class is really only useful for subclassing.
    """

    # ``ARG_DATA`` follows the same format as a member of ``ARG_TABLE`` in
    # ``BasicCommand`` class as specified in
    # ``awscli/customizations/commands.py``.
    #
    # For example, an ``ARG_DATA`` variable would be filled out as:
    #
    # ARG_DATA =
    # {'name': 'my-argument',
    #  'help_text': 'This is argument ensures the argument is specified'
    #               'no other arguments are required'}
    ARG_DATA = {'name': 'no-required-args'}

    def __init__(self, session):
        self._session = session
        self._register_argument_action()
        super(OverrideRequiredArgsArgument, self).__init__(**self.ARG_DATA)

    def _register_argument_action(self):
        self._session.register(
            'before-building-argument-table-parser',
            self.override_required_args,
        )

    def override_required_args(self, argument_table, args, **kwargs):
        name_in_cmdline = '--' + self.name
        # Set all ``Argument`` objects in ``argument_table`` to not required
        # if this argument's name is present in the command line.
        if name_in_cmdline in args:
            for arg_name in argument_table.keys():
                argument_table[arg_name].required = False


class StatefulArgument(CustomArgument):
    """An argument that maintains a stateful value"""

    def __init__(self, *args, **kwargs):
        super(StatefulArgument, self).__init__(*args, **kwargs)
        self._value = None

    def add_to_params(self, parameters, value):
        super(StatefulArgument, self).add_to_params(parameters, value)
        self._value = value

    @property
    def value(self):
        return self._value


class QueryOutFileArgument(StatefulArgument):
    """An argument that write a JMESPath query result to a file"""

    def __init__(
        self, session, name, query, after_call_event, perm, *args, **kwargs
    ):
        self._session = session
        self._query = query
        self._after_call_event = after_call_event
        self._perm = perm
        # Generate default help_text if text was not provided.
        if 'help_text' not in kwargs:
            kwargs['help_text'] = (
                'Saves the command output contents of %s '
                'to the given filename' % self.query
            )
        super(QueryOutFileArgument, self).__init__(name, *args, **kwargs)

    @property
    def query(self):
        return self._query

    @property
    def perm(self):
        return self._perm

    def add_to_params(self, parameters, value):
        value = resolve_given_outfile_path(value)
        super(QueryOutFileArgument, self).add_to_params(parameters, value)
        if self.value is not None:
            # Only register the event to save the argument if it is set
            self._session.register(self._after_call_event, self.save_query)

    def save_query(self, parsed, **kwargs):
        """Saves the result of a JMESPath expression to a file.

        This method only saves the query data if the response code of
        the parsed result is < 300.
        """
        if is_parsed_result_successful(parsed):
            contents = jmespath.search(self.query, parsed)
            with compat_open(
                self.value, 'w', access_permissions=self.perm
            ) as fp:
                # Don't write 'None' to a file -- write ''.
                if contents is None:
                    fp.write('')
                else:
                    fp.write(contents)
                # Even though the file is opened using the requested mode
                # (e.g. 0o600), the mode is only applied if a new file is
                # created. This means if the file already exists, its
                # permissions will not be changed. So, the os.chmod call is
                # retained here to preserve behavior of this argument always
                # clobbering a preexisting file's permissions to the desired
                # mode.
                os.chmod(self.value, self.perm)


class NestedBlobArgumentHoister:
    """Can be registered to update a single argument / model value combination
    mapping that to a new top-level argument.
    Currently limited to blob argument types as these are the only ones
    requiring the hoist.
    """

    def __init__(
        self,
        source_arg,
        source_arg_blob_member,
        new_arg,
        new_arg_doc_string,
        doc_string_addendum,
    ):
        self._source_arg = source_arg
        self._source_arg_blob_member = source_arg_blob_member
        self._new_arg = new_arg
        self._new_arg_doc_string = new_arg_doc_string
        self._doc_string_addendum = doc_string_addendum

    def __call__(self, session, argument_table, **kwargs):
        if not self._valid_target(argument_table):
            return
        self._update_arg(argument_table, self._source_arg, self._new_arg)

    def _valid_target(self, argument_table):
        # Find the source argument and check that it has a member of
        # the same name and type.
        if self._source_arg in argument_table:
            arg = argument_table[self._source_arg]
            input_model = arg.argument_model
            member = input_model.members.get(self._source_arg_blob_member)
            if member is not None and member.type_name == 'blob':
                return True
        return False

    def _update_arg(self, argument_table, source_arg, new_arg):
        argument_table[new_arg] = _NestedBlobArgumentParamOverwrite(
            new_arg,
            source_arg,
            self._source_arg_blob_member,
            help_text=self._new_arg_doc_string,
            cli_type_name='blob',
        )
        argument_table[source_arg].required = False
        argument_table[source_arg].documentation += self._doc_string_addendum


class _NestedBlobArgumentParamOverwrite(CustomArgument):
    def __init__(self, new_arg, source_arg, source_arg_blob_member, **kwargs):
        super(_NestedBlobArgumentParamOverwrite, self).__init__(
            new_arg, **kwargs
        )
        self._param_to_overwrite = _reverse_xform_name(source_arg)
        self._source_arg_blob_member = source_arg_blob_member

    def add_to_params(self, parameters, value):
        if value is None:
            return
        param_value = {self._source_arg_blob_member: value}
        if parameters.get(self._param_to_overwrite):
            parameters[self._param_to_overwrite].update(param_value)
        else:
            parameters[self._param_to_overwrite] = param_value


def _upper(match):
    return match.group(1).lstrip('-').upper()


def _reverse_xform_name(name):
    return re.sub(r'(^.|-.)', _upper, name)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

"""Service view collector for ECS Express Gateway Service monitoring.

This module provides business logic for collecting and formatting
ECS Express Gateway Service monitoring data.
"""

import time
from functools import reduce

from botocore.exceptions import ClientError

from awscli.customizations.ecs.exceptions import MonitoringError
from awscli.customizations.ecs.expressgateway.managedresource import (
    ManagedResource,
)
from awscli.customizations.ecs.expressgateway.managedresourcegroup import (
    ManagedResourceGroup,
)


class ServiceViewCollector:
    """Collects and formats ECS Express Gateway Service monitoring data.

    Responsible for:
    - Making ECS API calls
    - Parsing resource data
    - Formatting output strings
    - Caching responses

    Args:
        client: ECS client for API calls
        service_arn (str): ARN of the service to monitor
        mode (str): Monitoring mode - 'RESOURCE' or 'DEPLOYMENT'
        use_color (bool): Whether to use color in output
    """

    def __init__(self, client, service_arn, mode, use_color=True):
        self._client = client
        self.service_arn = service_arn
        self.mode = mode
        self.use_color = use_color
        self.last_described_gateway_service_response = None
        self.last_execution_time = 0
        self.cached_monitor_result = None

    def get_current_view(
        self, spinner_char="{SPINNER}", execution_refresh_millis=5000
    ):
        """Get current monitoring view as formatted string.

        Args:
            spinner_char (str): Character for progress indication
            execution_refresh_millis (int): Refresh interval in milliseconds

        Returns:
            str: Formatted monitoring output
        """
        current_time = time.time()

        if (
            current_time - self.last_execution_time
            >= execution_refresh_millis / 1000.0
        ):
            try:
                describe_gateway_service_response = (
                    self._client.describe_express_gateway_service(
                        serviceArn=self.service_arn
                    )
                )
                if not describe_gateway_service_response:
                    self.cached_monitor_result = (
                        None,
                        "Trying to describe gateway service",
                    )
                elif (
                    not (
                        service := describe_gateway_service_response.get(
                            "service"
                        )
                    )
                    or not service.get("serviceArn")
                    or service.get("activeConfigurations") is None
                ):
                    self.cached_monitor_result = (
                        None,
                        "Trying to describe gateway service",
                    )
                else:
                    self.last_described_gateway_service_response = (
                        describe_gateway_service_response
                    )
                    described_gateway_service = (
                        describe_gateway_service_response.get("service")
                    )

                    if self.mode == "DEPLOYMENT":
                        managed_resources, info = self._diff_service_view(
                            described_gateway_service
                        )
                    else:
                        managed_resources, info = self._combined_service_view(
                            described_gateway_service
                        )

                    service_resources = [
                        self._parse_cluster(described_gateway_service),
                        self._parse_service(described_gateway_service),
                    ]
                    if managed_resources:
                        service_resources.append(managed_resources)
                    service_resource = ManagedResourceGroup(
                        resources=service_resources
                    )
                    self._update_cached_monitor_results(service_resource, info)
            except ClientError as e:
                if (
                    e.response.get('Error', {}).get('Code')
                    == 'InvalidParameterException'
                ):
                    error_message = e.response.get('Error', {}).get(
                        'Message', ''
                    )
                    if (
                        "Cannot call DescribeServiceRevisions for a service that is INACTIVE"
                        in error_message
                    ):
                        empty_resource_group = ManagedResourceGroup()
                        self._update_cached_monitor_results(
                            empty_resource_group, "Service is inactive"
                        )
                    else:
                        raise
                else:
                    raise

            self.last_execution_time = current_time

        if not self.cached_monitor_result:
            return "Waiting for initial data"
        else:
            service_resource, info = self.cached_monitor_result
            status_string = (
                service_resource.get_status_string(
                    spinner_char=spinner_char, use_color=self.use_color
                )
                if service_resource
                else None
            )

            output = "\n".join([x for x in [status_string, info] if x])
            return output

    def _diff_service_view(self, describe_gateway_service_response):
        """Generate diff view showing changes in the latest deployment."""
        service_arn = describe_gateway_service_response.get("serviceArn")
        list_service_deployments_response = (
            self._client.list_service_deployments(
                service=service_arn, maxResults=1
            )
        )
        listed_service_deployments = self._validate_and_parse_response(
            list_service_deployments_response,
            "ListServiceDeployments",
            expected_field="serviceDeployments",
        )
        if (
            not listed_service_deployments
            or "serviceDeploymentArn" not in listed_service_deployments[0]
        ):
            return (
                None,
                "Waiting for a deployment to start",
            )

        deployment_arn = listed_service_deployments[0].get(
            "serviceDeploymentArn"
        )

        describe_service_deployments_response = (
            self._client.describe_service_deployments(
                serviceDeploymentArns=[deployment_arn]
            )
        )
        described_service_deployments = self._validate_and_parse_response(
            describe_service_deployments_response,
            "DescribeServiceDeployments",
            expected_field="serviceDeployments",
            eventually_consistent=True,
        )
        if not described_service_deployments:
            return (None, "Waiting for a deployment to start")

        described_service_deployment = described_service_deployments[0]
        if (
            not described_service_deployment
            or not described_service_deployment.get("targetServiceRevision")
        ):
            return (
                None,
                "Waiting for a deployment to start",
            )

        target_sr = described_service_deployment.get(
            "targetServiceRevision"
        ).get("arn")

        target_sr_resources_list, described_target_sr_list = (
            self._describe_and_parse_service_revisions([target_sr])
        )
        if len(target_sr_resources_list) != 1:
            return (None, "Trying to describe service revisions")
        target_sr_resources = target_sr_resources_list[0]
        described_target_sr = described_target_sr_list[0]

        task_def_arn = described_target_sr.get("taskDefinition")
        if "sourceServiceRevisions" in described_service_deployment:
            source_sr_resources, _ = (
                self._describe_and_parse_service_revisions(
                    [
                        sr.get("arn")
                        for sr in described_service_deployment.get(
                            "sourceServiceRevisions"
                        )
                    ]
                )
            )
            if len(source_sr_resources) != len(
                described_service_deployment.get("sourceServiceRevisions")
            ):
                return (None, "Trying to describe service revisions")
            source_sr_resources_combined = reduce(
                lambda x, y: x.combine(y), source_sr_resources
            )
        else:
            source_sr_resources_combined = ManagedResourceGroup()

        updating_resources, disassociating_resources = (
            target_sr_resources.compare_resource_sets(
                source_sr_resources_combined
            )
        )
        updating_resources.resource_type = "Updating"
        disassociating_resources.resource_type = "Disassociating"
        service_resources = ManagedResourceGroup(
            resource_type="Deployment",
            identifier=deployment_arn,
            status=described_service_deployment.get("status"),
            reason=described_service_deployment.get("statusReason"),
            resources=[
                ManagedResource(
                    resource_type="TargetServiceRevision", identifier=target_sr
                ),
                ManagedResource(
                    resource_type="TaskDefinition", identifier=task_def_arn
                ),
                updating_resources,
                disassociating_resources,
            ],
        )
        return service_resources, None

    def _combined_service_view(self, describe_gateway_service_response):
        """Generate combined view of all active service resources."""
        service_revision_arns = [
            config.get("serviceRevisionArn")
            for config in describe_gateway_service_response.get(
                "activeConfigurations"
            )
        ]
        service_revision_resources, _ = (
            self._describe_and_parse_service_revisions(service_revision_arns)
        )

        # If empty, we're still waiting for active configurations
        if not service_revision_resources or len(
            service_revision_resources
        ) != len(service_revision_arns):
            return (None, "Trying to describe service revisions")

        service_resource = reduce(
            lambda x, y: x.combine(y), service_revision_resources
        )

        return service_resource, None

    def _update_cached_monitor_results(self, resource, info):
        """Update cached monitoring results with new data."""
        if not self.cached_monitor_result:
            self.cached_monitor_result = (resource, info)
        else:
            self.cached_monitor_result = (
                resource or self.cached_monitor_result[0],
                info,
            )

    def _validate_and_parse_response(
        self,
        response,
        operation_name,
        expected_field=None,
        eventually_consistent=False,
    ):
        """Validate API response and extract expected field."""
        if not response:
            raise MonitoringError(f"{operation_name} response is empty")

        self._parse_failures(response, operation_name, eventually_consistent)

        if not expected_field:
            return None

        if response.get(expected_field) is None:
            raise MonitoringError(
                f"{operation_name} response is missing {expected_field}"
            )
        return response.get(expected_field)

    def _parse_failures(self, response, operation_name, eventually_consistent):
        """Parse and raise errors for API response failures."""
        failures = response.get("failures")

        if not failures:
            return

        if any(not f.get('arn') or not f.get('reason') for f in failures):
            raise MonitoringError(
                "Invalid failure response: missing arn or reason"
            )

        if eventually_consistent:
            failures = [
                failure
                for failure in failures
                if failure.get("reason") != "MISSING"
            ]

        if not failures:
            return

        failure_msgs = [
            f"{f['arn']} failed with {f['reason']}" for f in failures
        ]
        joined_msgs = '\n'.join(failure_msgs)
        raise MonitoringError(f"{operation_name}:\n{joined_msgs}")

    def _describe_and_parse_service_revisions(self, arns):
        """Describe and parse service revisions into managed resources."""
        describe_service_revisions_response = (
            self._client.describe_service_revisions(serviceRevisionArns=arns)
        )
        described_service_revisions = self._validate_and_parse_response(
            describe_service_revisions_response,
            "DescribeServiceRevisions",
            expected_field="serviceRevisions",
            eventually_consistent=True,
        )

        return [
            self._parse_ecs_managed_resources(sr)
            for sr in described_service_revisions
        ], described_service_revisions

    def _parse_cluster(self, service):
        return ManagedResource("Cluster", service.get("cluster"))

    def _parse_service(self, service):
        service_arn = service.get("serviceArn")
        cluster = service.get("cluster")
        describe_service_response = self._client.describe_services(
            cluster=cluster, services=[service_arn]
        )
        described_service = self._validate_and_parse_response(
            describe_service_response, "DescribeServices", "services"
        )[0]
        return ManagedResource(
            "Service",
            service.get("serviceArn"),
            additional_info=described_service
            and described_service.get("events")[0].get("message")
            if described_service.get("events")
            else None,
        )

    def _parse_ecs_managed_resources(self, service_revision):
        managed_resources = service_revision.get("ecsManagedResources")
        if not managed_resources:
            return ManagedResourceGroup()

        parsed_resources = []
        if "ingressPaths" in managed_resources:
            parsed_resources.append(
                ManagedResourceGroup(
                    resource_type="IngressPaths",
                    resources=[
                        self._parse_ingress_path_resources(ingress_path)
                        for ingress_path in managed_resources.get(
                            "ingressPaths"
                        )
                    ],
                )
            )
        if "autoScaling" in managed_resources:
            parsed_resources.append(
                self._parse_auto_scaling_configuration(
                    managed_resources.get("autoScaling")
                )
            )
        if "metricAlarms" in managed_resources:
            parsed_resources.append(
                self._parse_metric_alarms(
                    managed_resources.get("metricAlarms")
                )
            )
        if "serviceSecurityGroups" in managed_resources:
            parsed_resources.append(
                self._parse_service_security_groups(
                    managed_resources.get("serviceSecurityGroups")
                )
            )
        if "logGroups" in managed_resources:
            parsed_resources.append(
                self._parse_log_groups(managed_resources.get("logGroups"))
            )
        return ManagedResourceGroup(resources=parsed_resources)

    def _parse_ingress_path_resources(self, ingress_path):
        resources = []
        if ingress_path.get("loadBalancer"):
            resources.append(
                self._parse_managed_resource(
                    ingress_path.get("loadBalancer"), "LoadBalancer"
                )
            )
        if ingress_path.get("loadBalancerSecurityGroups"):
            resources.extend(
                self._parse_managed_resource_list(
                    ingress_path.get("loadBalancerSecurityGroups"),
                    "LoadBalancerSecurityGroup",
                )
            )
        if ingress_path.get("certificate"):
            resources.append(
                self._parse_managed_resource(
                    ingress_path.get("certificate"), "Certificate"
                )
            )
        if ingress_path.get("listener"):
            resources.append(
                self._parse_managed_resource(
                    ingress_path.get("listener"), "Listener"
                )
            )
        if ingress_path.get("rule"):
            resources.append(
                self._parse_managed_resource(ingress_path.get("rule"), "Rule")
            )
        if ingress_path.get("targetGroups"):
            resources.extend(
                self._parse_managed_resource_list(
                    ingress_path.get("targetGroups"), "TargetGroup"
                )
            )
        return ManagedResourceGroup(
            resource_type="IngressPath",
            identifier=ingress_path.get("endpoint"),
            resources=resources,
        )

    def _parse_auto_scaling_configuration(self, auto_scaling_configuration):
        resources = []
        if auto_scaling_configuration.get("scalableTarget"):
            resources.append(
                self._parse_managed_resource(
                    auto_scaling_configuration.get("scalableTarget"),
                    "ScalableTarget",
                )
            )
        if auto_scaling_configuration.get("applicationAutoScalingPolicies"):
            resources.extend(
                self._parse_managed_resource_list(
                    auto_scaling_configuration.get(
                        "applicationAutoScalingPolicies"
                    ),
                    "ApplicationAutoScalingPolicy",
                )
            )
        return ManagedResourceGroup(
            resource_type="AutoScalingConfiguration", resources=resources
        )

    def _parse_metric_alarms(self, metric_alarms):
        return ManagedResourceGroup(
            resource_type="MetricAlarms",
            resources=self._parse_managed_resource_list(
                metric_alarms, "MetricAlarm"
            ),
        )

    def _parse_service_security_groups(self, service_security_groups):
        return ManagedResourceGroup(
            resource_type="ServiceSecurityGroups",
            resources=self._parse_managed_resource_list(
                service_security_groups, "SecurityGroup"
            ),
        )

    def _parse_log_groups(self, logs_groups):
        return ManagedResourceGroup(
            resource_type="LogGroups",
            resources=self._parse_managed_resource_list(
                logs_groups, "LogGroup"
            ),
        )

    def _parse_managed_resource(self, resource, resource_type):
        return ManagedResource(
            resource_type,
            resource.get("arn"),
            status=resource.get("status"),
            updated_at=resource.get("updatedAt"),
            reason=resource.get("statusReason"),
        )

    def _parse_managed_resource_list(self, data_list, resource_type):
        return [
            self._parse_managed_resource(data, resource_type)
            for data in data_list
        ]



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import base64
import logging
import os

from awscrt.crypto import RSA, RSASignatureAlgorithm
from botocore import model

from awscli.arguments import BaseCLIArgument

logger = logging.getLogger(__name__)


HELP = """<p>The file that contains the private key used to launch
the instance (e.g. windows-keypair.pem).  If this is supplied, the
password data sent from EC2 will be decrypted before display.</p>"""


def ec2_add_priv_launch_key(
    argument_table, operation_model, session, **kwargs
):
    """
    This handler gets called after the argument table for the
    operation has been created.  It's job is to add the
    ``priv-launch-key`` parameter.
    """
    argument_table['priv-launch-key'] = LaunchKeyArgument(
        session, operation_model, 'priv-launch-key'
    )


class LaunchKeyArgument(BaseCLIArgument):
    def __init__(self, session, operation_model, name):
        self._session = session
        self.argument_model = model.Shape(
            'LaunchKeyArgument', {'type': 'string'}
        )
        self._operation_model = operation_model
        self._name = name
        self._key_path = None
        self._required = False

    @property
    def cli_type_name(self):
        return 'string'

    @property
    def required(self):
        return self._required

    @required.setter
    def required(self, value):
        self._required = value

    @property
    def documentation(self):
        return HELP

    def add_to_parser(self, parser):
        parser.add_argument(
            self.cli_name, dest=self.py_name, help='SSH Private Key file'
        )

    def add_to_params(self, parameters, value):
        """
        This gets called with the value of our ``--priv-launch-key``
        if it is specified.  It needs to determine if the path
        provided is valid and, if it is, it stores it in the instance
        variable ``_key_path`` for use by the decrypt routine.
        """
        if value:
            path = os.path.expandvars(value)
            path = os.path.expanduser(path)
            if os.path.isfile(path):
                self._key_path = path
                service_id = self._operation_model.service_model.service_id
                event = 'after-call.%s.%s' % (
                    service_id.hyphenize(),
                    self._operation_model.name,
                )
                self._session.register(event, self._decrypt_password_data)
            else:
                msg = (
                    'priv-launch-key should be a path to the '
                    'local SSH private key file used to launch '
                    'the instance.'
                )
                raise ValueError(msg)

    def _decrypt_password_data(self, parsed, **kwargs):
        """
        This handler gets called after the GetPasswordData command has been
        executed. It is called with the ``parsed`` data. It checks to
        see if a private launch key was specified on the command. If it was,
        it tries to use that private key to decrypt the password data and
        replace it in the returned data dictionary.
        """
        if self._key_path is not None:
            logger.debug("Decrypting password data using: %s", self._key_path)
            value = parsed.get('PasswordData')
            if not value:
                return
            try:
                with open(self._key_path, 'rb') as pk_file:
                    pk_bytes = pk_file.read()
                    private_key = RSA.new_private_key_from_pem_data(pk_bytes)
                    value = base64.b64decode(value)
                    value = private_key.decrypt(
                        RSASignatureAlgorithm.PKCS1_5_SHA256, value
                    )
                    logger.debug(parsed)
                    parsed['PasswordData'] = value.decode('utf-8')
                    logger.debug(parsed)
            except Exception:
                logger.debug('Unable to decrypt PasswordData', exc_info=True)
                msg = (
                    'Unable to decrypt password data using '
                    'provided private key file.'
                )
                raise ValueError(msg)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from .parser import Parser


class AttributeExtractor:
    COMPARATORS = {
        'eq': '=',
        'ne': '<>',
        'lt': '<',
        'lte': '<=',
        'gt': '>',
        'gte': '>=',
    }

    def __init__(self, parser=None):
        self._parser = parser
        if parser is None:
            self._parser = Parser()
        self._identifiers = {}
        self._literals = {}
        self._index_offset = 0

    def extract(self, expression, index_offset=0):
        self._identifiers = {}
        self._literals = {}
        self._index_offset = index_offset
        root = self._parser.parse(expression)
        expression = self._visit(root).strip()
        return {
            'expression': expression,
            'identifiers': self._identifiers,
            'values': self._literals,
            'substitution_count': len(self._identifiers) + len(self._literals),
        }

    def _substitution_index(self):
        num_substituted = len(self._identifiers) + len(self._literals)
        return num_substituted + self._index_offset

    def _visit(self, node):
        method = getattr(self, '_visit_%s' % node['type'])
        return method(node)

    def _visit_comparator(self, node):
        left = self._visit(node['children'][0])
        right = self._visit(node['children'][1])
        expression = '%s%s %s' % (left, self.COMPARATORS[node['value']], right)
        return expression

    def _visit_identifier(self, node):
        identifier_replacement = '#n%s' % self._substitution_index()
        self._identifiers[identifier_replacement] = node['value']
        return '%s ' % identifier_replacement

    def _visit_path_identifier(self, node):
        left = self._visit(node['children'][0]).strip()
        right = self._visit(node['children'][1])
        return '%s.%s' % (left, right)

    def _visit_index_identifier(self, node):
        left = self._visit(node['children'][0]).strip()
        return '%s[%s] ' % (left, node['value'])

    def _visit_literal(self, node):
        literal_replacement = ':n%s' % self._substitution_index()
        self._literals[literal_replacement] = node['value']
        return '%s ' % literal_replacement

    def _visit_sequence(self, node):
        visited_children = []
        for child in node['children']:
            visited_children.append(self._visit(child).strip())
        return '%s' % ', '.join(visited_children)

    def _visit_or_expression(self, node):
        left = self._visit(node['children'][0])
        right = self._visit(node['children'][1])
        return '%sOR %s' % (left, right)

    def _visit_and_expression(self, node):
        left = self._visit(node['children'][0])
        right = self._visit(node['children'][1])
        return '%sAND %s' % (left, right)

    def _visit_not_expression(self, node):
        return 'NOT %s' % self._visit(node['children'][0])

    def _visit_subexpression(self, node):
        return '( %s) ' % self._visit(node['children'][0])

    def _visit_function(self, node):
        return '%s(%s) ' % (node['value'], self._visit_sequence(node).strip())

    def _visit_in_expression(self, node):
        return '%sIN (%s) ' % (
            self._visit(node['children'][0]),
            self._visit(node['children'][1]).strip(),
        )

    def _visit_between_expression(self, node):
        return '%sBETWEEN %sAND %s' % (
            self._visit(node['children'][0]),
            self._visit(node['children'][1]),
            self._visit(node['children'][2]),
        )



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import sys
from base64 import b64decode

from awscli.customizations.commands import BasicCommand
from awscli.customizations.utils import create_client_from_parsed_globals


def register_ecr_commands(cli):
    cli.register('building-command-table.ecr', _inject_commands)


def _inject_commands(command_table, session, **kwargs):
    command_table['get-login-password'] = ECRGetLoginPassword(session)


class ECRGetLoginPassword(BasicCommand):
    """Get a password to be used with container clients such as Docker"""

    NAME = 'get-login-password'

    DESCRIPTION = BasicCommand.FROM_FILE(
        'ecr/get-login-password_description.rst'
    )

    def _run_main(self, parsed_args, parsed_globals):
        ecr_client = create_client_from_parsed_globals(
            self._session, 'ecr', parsed_globals
        )
        result = ecr_client.get_authorization_token()
        auth = result['authorizationData'][0]
        auth_token = b64decode(auth['authorizationToken']).decode()
        _, password = auth_token.split(':')
        sys.stdout.write(password)
        sys.stdout.write('\n')
        return 0



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import inspect
import os

from botocore import xform_name
from botocore.compat import OrderedDict
from botocore.docs.bcdoc.restdoc import DocumentStructure
from botocore.docs.example import ResponseExampleDocumenter
from botocore.docs.method import (
    document_custom_method,
    document_model_driven_method,
    get_instance_public_methods,
)
from botocore.docs.params import ResponseParamsDocumenter
from botocore.docs.sharedexample import document_shared_examples
from botocore.docs.utils import DocumentedShape, get_official_service_name


def _allowlist_generate_presigned_url(method_name, service_name, **kwargs):
    if method_name != 'generate_presigned_url':
        return None
    return service_name in ['s3']


class ClientDocumenter:
    _CLIENT_METHODS_FILTERS = [
        _allowlist_generate_presigned_url,
    ]

    def __init__(self, client, root_docs_path, shared_examples=None):
        self._client = client
        self._client_class_name = self._client.__class__.__name__
        self._root_docs_path = root_docs_path
        self._shared_examples = shared_examples
        if self._shared_examples is None:
            self._shared_examples = {}
        self._service_name = self._client.meta.service_model.service_name

    def document_client(self, section):
        """Documents a client and its methods

        :param section: The section to write to.
        """
        self._add_title(section)
        self._add_class_signature(section)
        client_methods = self._get_client_methods()
        self._add_client_intro(section, client_methods)
        self._add_client_methods(client_methods)

    def _get_client_methods(self):
        client_methods = get_instance_public_methods(self._client)
        return self._filter_client_methods(client_methods)

    def _filter_client_methods(self, client_methods):
        filtered_methods = {}
        for method_name, method in client_methods.items():
            include = self._filter_client_method(
                method=method,
                method_name=method_name,
                service_name=self._service_name,
            )
            if include:
                filtered_methods[method_name] = method
        return filtered_methods

    def _filter_client_method(self, **kwargs):
        for filter in self._CLIENT_METHODS_FILTERS:
            filter_include = filter(**kwargs)
            if filter_include is not None:
                return filter_include
        return True

    def _add_title(self, section):
        section.style.h2('Client')

    def _add_client_intro(self, section, client_methods):
        section = section.add_new_section('intro')
        # Write out the top level description for the client.
        official_service_name = get_official_service_name(
            self._client.meta.service_model
        )
        section.write(
            f'A low-level client representing {official_service_name}'
        )
        section.style.new_line()
        section.include_doc_string(
            self._client.meta.service_model.documentation
        )

        # Write out the client example instantiation.
        self._add_client_creation_example(section)

        # List out all of the possible client methods.
        section.style.dedent()
        section.style.new_paragraph()
        section.writeln('These are the available methods:')
        section.style.toctree()
        for method_name in sorted(client_methods):
            section.style.tocitem(f'{self._service_name}/client/{method_name}')

    def _add_class_signature(self, section):
        section.style.start_sphinx_py_class(
            class_name=f'{self._client_class_name}.Client'
        )

    def _add_client_creation_example(self, section):
        section.style.start_codeblock()
        section.style.new_line()
        section.write(
            f'client = session.create_client(\'{self._service_name}\')'
        )
        section.style.end_codeblock()

    def _add_client_methods(self, client_methods):
        for method_name in sorted(client_methods):
            # Create a new DocumentStructure for each client method and add contents.
            method_doc_structure = DocumentStructure(
                method_name, target='html'
            )
            self._add_client_method(
                method_doc_structure, method_name, client_methods[method_name]
            )
            # Write client methods in individual/nested files.
            # Path: <root>/reference/services/<service>/client/<method_name>.rst
            client_dir_path = os.path.join(
                self._root_docs_path, self._service_name, 'client'
            )
            method_doc_structure.write_to_file(client_dir_path, method_name)

    def _add_client_method(self, section, method_name, method):
        section.add_title_section(method_name)
        method_section = section.add_new_section(method_name)
        if self._is_custom_method(method_name):
            self._add_custom_method(method_section, method_name, method)
        else:
            self._add_model_driven_method(method_section, method_name)

    def _is_custom_method(self, method_name):
        return method_name not in self._client.meta.method_to_api_mapping

    def _add_custom_method(self, section, method_name, method):
        document_custom_method(section, method_name, method)

    def _add_method_exceptions_list(self, section, operation_model):
        error_section = section.add_new_section('exceptions')
        error_section.style.new_line()
        error_section.style.bold('Exceptions')
        error_section.style.new_line()
        for error in operation_model.error_shapes:
            class_name = (
                f'{self._client_class_name}.Client.exceptions.{error.name}'
            )
            error_section.style.li(f':py:class:`{class_name}`')

    def _add_model_driven_method(self, section, method_name):
        service_model = self._client.meta.service_model
        operation_name = self._client.meta.method_to_api_mapping[method_name]
        operation_model = service_model.operation_model(operation_name)

        example_prefix = f'response = client.{method_name}'
        document_model_driven_method(
            section,
            method_name,
            operation_model,
            event_emitter=self._client.meta.events,
            method_description=operation_model.documentation,
            example_prefix=example_prefix,
        )

        # Add any modeled exceptions
        if operation_model.error_shapes:
            self._add_method_exceptions_list(section, operation_model)

        # Add the shared examples
        shared_examples = self._shared_examples.get(operation_name)
        if shared_examples:
            document_shared_examples(
                section, operation_model, example_prefix, shared_examples
            )


class ClientExceptionsDocumenter:
    _USER_GUIDE_LINK = (
        'https://boto3.amazonaws.com/'
        'v1/documentation/api/latest/guide/error-handling.html'
    )
    _GENERIC_ERROR_SHAPE = DocumentedShape(
        name='Error',
        type_name='structure',
        documentation=('Normalized access to common exception attributes.'),
        members=OrderedDict(
            [
                (
                    'Code',
                    DocumentedShape(
                        name='Code',
                        type_name='string',
                        documentation=(
                            'An identifier specifying the exception type.'
                        ),
                    ),
                ),
                (
                    'Message',
                    DocumentedShape(
                        name='Message',
                        type_name='string',
                        documentation=(
                            'A descriptive message explaining why the exception '
                            'occured.'
                        ),
                    ),
                ),
            ]
        ),
    )

    def __init__(self, client, root_docs_path):
        self._client = client
        self._client_class_name = self._client.__class__.__name__
        self._service_name = self._client.meta.service_model.service_name
        self._service_id = self._client.meta.service_model.service_id
        self._root_docs_path = root_docs_path

    def document_exceptions(self, section):
        self._add_title(section)
        self._add_overview(section)
        self._add_exceptions_list(section)
        self._add_exception_classes()

    def _add_title(self, section):
        section.style.h2('Client Exceptions')

    def _add_overview(self, section):
        section.style.new_line()
        section.write(
            'Client exceptions are available on a client instance '
            'via the ``exceptions`` property. For more detailed instructions '
            'and examples on the exact usage of client exceptions, see the '
            'error handling '
        )
        section.style.external_link(
            title='user guide',
            link=self._USER_GUIDE_LINK,
        )
        section.write('.')
        section.style.new_line()

    def _exception_class_name(self, shape):
        cls_name = self._client.__class__.__name__
        return f'{self._client_class_name}.Client.exceptions.{shape.name}'

    def _add_exceptions_list(self, section):
        error_shapes = self._client.meta.service_model.error_shapes
        if not error_shapes:
            section.style.new_line()
            section.write('This client has no modeled exception classes.')
            section.style.new_line()
            return
        section.style.new_line()
        section.writeln('The available client exceptions are:')
        section.style.toctree()
        for shape in error_shapes:
            section.style.tocitem(
                f'{self._service_name}/client/exceptions/{shape.name}'
            )

    def _add_exception_classes(self):
        for shape in self._client.meta.service_model.error_shapes:
            # Create a new DocumentStructure for each exception method and add contents.
            exception_doc_structure = DocumentStructure(
                shape.name, target='html'
            )
            self._add_exception_class(exception_doc_structure, shape)
            # Write exceptions in individual/nested files.
            # Path: <root>/reference/services/<service>/client/exceptions/<exception_name>.rst
            exception_dir_path = os.path.join(
                self._root_docs_path,
                self._service_name,
                'client',
                'exceptions',
            )
            exception_doc_structure.write_to_file(
                exception_dir_path, shape.name
            )

    def _add_exception_class(self, section, shape):
        section.add_title_section(shape.name)
        class_section = section.add_new_section(shape.name)
        class_name = self._exception_class_name(shape)
        class_section.style.start_sphinx_py_class(class_name=class_name)
        self._add_top_level_documentation(class_section, shape)
        self._add_exception_catch_example(class_section, shape)
        self._add_response_attr(class_section, shape)
        class_section.style.end_sphinx_py_class()

    def _add_top_level_documentation(self, section, shape):
        if shape.documentation:
            section.style.new_line()
            section.include_doc_string(shape.documentation)
            section.style.new_line()

    def _add_exception_catch_example(self, section, shape):
        section.style.new_line()
        section.style.bold('Example')
        section.style.start_codeblock()
        section.write('try:')
        section.style.indent()
        section.style.new_line()
        section.write('...')
        section.style.dedent()
        section.style.new_line()
        section.write(f'except client.exceptions.{shape.name} as e:')
        section.style.indent()
        section.style.new_line()
        section.write('print(e.response)')
        section.style.dedent()
        section.style.end_codeblock()

    def _add_response_attr(self, section, shape):
        response_section = section.add_new_section('response')
        response_section.style.start_sphinx_py_attr('response')
        self._add_response_attr_description(response_section)
        self._add_response_example(response_section, shape)
        self._add_response_params(response_section, shape)
        response_section.style.end_sphinx_py_attr()

    def _add_response_attr_description(self, section):
        section.style.new_line()
        section.include_doc_string(
            'The parsed error response. All exceptions have a top level '
            '``Error`` key that provides normalized access to common '
            'exception atrributes. All other keys are specific to this '
            'service or exception class.'
        )
        section.style.new_line()

    def _add_response_example(self, section, shape):
        example_section = section.add_new_section('syntax')
        example_section.style.new_line()
        example_section.style.bold('Syntax')
        example_section.style.new_paragraph()
        documenter = ResponseExampleDocumenter(
            service_id=self._service_id,
            operation_name=None,
            event_emitter=self._client.meta.events,
        )
        documenter.document_example(
            example_section,
            shape,
            include=[self._GENERIC_ERROR_SHAPE],
        )

    def _add_response_params(self, section, shape):
        params_section = section.add_new_section('Structure')
        params_section.style.new_line()
        params_section.style.bold('Structure')
        params_section.style.new_paragraph()
        documenter = ResponseParamsDocumenter(
            service_id=self._service_id,
            operation_name=None,
            event_emitter=self._client.meta.events,
        )
        documenter.document_params(
            params_section,
            shape,
            include=[self._GENERIC_ERROR_SHAPE],
        )


class ClientContextParamsDocumenter:
    _CONFIG_GUIDE_LINK = (
        'https://boto3.amazonaws.com/'
        'v1/documentation/api/latest/guide/configuration.html'
    )

    OMITTED_CONTEXT_PARAMS = {
        's3': (
            'Accelerate',
            'DisableMultiRegionAccessPoints',
            'ForcePathStyle',
            'UseArnRegion',
        ),
        's3control': ('UseArnRegion',),
    }

    def __init__(self, service_name, context_params):
        self._service_name = service_name
        self._context_params = context_params

    def document_context_params(self, section):
        self._add_title(section)
        self._add_overview(section)
        self._add_context_params_list(section)

    def _add_title(self, section):
        section.style.h2('Client Context Parameters')

    def _add_overview(self, section):
        section.style.new_line()
        section.write(
            'Client context parameters are configurable on a client '
            'instance via the ``client_context_params`` parameter in the '
            '``Config`` object. For more detailed instructions and examples '
            'on the exact usage of context params see the '
        )
        section.style.external_link(
            title='configuration guide',
            link=self._CONFIG_GUIDE_LINK,
        )
        section.write('.')
        section.style.new_line()

    def _add_context_params_list(self, section):
        section.style.new_line()
        sn = f'``{self._service_name}``'
        section.writeln(f'The available {sn} client context params are:')
        for param in self._context_params:
            section.style.new_line()
            name = f'``{xform_name(param.name)}``'
            section.write(f'* {name} ({param.type}) - {param.documentation}')



# Copyright 2012-2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from awscli.customizations.commands import BasicCommand
from awscli.customizations.s3uploader import S3Uploader
from awscli.customizations.servicecatalog import exceptions
from awscli.customizations.servicecatalog.utils import get_s3_path, make_url


class GenerateBaseCommand(BasicCommand):
    def _run_main(self, parsed_args, parsed_globals):
        self.region = self.get_and_validate_region(parsed_globals)
        self.s3_client = self._session.create_client(
            's3',
            region_name=self.region,
            endpoint_url=parsed_globals.endpoint_url,
            verify=parsed_globals.verify_ssl,
        )
        self.s3_uploader = S3Uploader(
            self.s3_client, parsed_args.bucket_name, force_upload=True
        )
        try:
            self.s3_uploader.upload(
                parsed_args.file_path, get_s3_path(parsed_args.file_path)
            )
        except OSError:
            raise RuntimeError("%s cannot be found" % parsed_args.file_path)

    def get_and_validate_region(self, parsed_globals):
        region = parsed_globals.region
        if region is None:
            region = self._session.get_config_variable('region')
        if region not in self._session.get_available_regions('servicecatalog'):
            raise exceptions.InvalidParametersException(
                message=f"Region {parsed_globals.region} is not supported"
            )
        return region

    def create_s3_url(self, bucket_name, file_path):
        return make_url(self.region, bucket_name, get_s3_path(file_path))



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import os

from s3transfer.manager import TransferManager

from awscli.compat import get_binary_stdin
from awscli.customizations.s3.results import (
    CommandResultRecorder,
    DoneResultSubscriber,
    DryRunResult,
    FailureResult,
    NoProgressResultPrinter,
    OnlyShowErrorsResultPrinter,
    ProgressResultSubscriber,
    QueuedResult,
    QueuedResultSubscriber,
    ResultPrinter,
    ResultProcessor,
    ResultRecorder,
    SuccessResult,
)
from awscli.customizations.s3.subscribers import (
    CaseConflictCleanupSubscriber,
    CopyPropsSubscriberFactory,
    DeleteCopySourceObjectSubscriber,
    DeleteSourceFileSubscriber,
    DeleteSourceObjectSubscriber,
    DirectoryCreatorSubscriber,
    ProvideETagSubscriber,
    ProvideLastModifiedTimeSubscriber,
    ProvideSizeSubscriber,
    ProvideUploadContentTypeSubscriber,
)
from awscli.customizations.s3.transferconfig import (
    create_transfer_config_from_runtime_config,
)
from awscli.customizations.s3.utils import (
    MAX_UPLOAD_SIZE,
    NonSeekableStream,
    RequestParamsMapper,
    StdoutBytesWriter,
    create_warning,
    find_bucket_key,
    human_readable_size,
    relative_path,
)

LOGGER = logging.getLogger(__name__)


class S3TransferHandlerFactory:
    def __init__(self, cli_params):
        """Factory for S3TransferHandlers

        :type cli_params: dict
        :param cli_params: The parameters provide to the CLI command
        """
        self._cli_params = cli_params

    def __call__(self, transfer_manager, result_queue):
        """Creates a S3TransferHandler instance

        :param transfer_manager: The transfer manager to power the
            S3TransferHandler

        :type result_queue: queue.Queue
        :param result_queue: The result queue to be used to process results
            for the S3TransferHandler

        :returns: A S3TransferHandler instance
        """
        result_recorder = ResultRecorder()
        result_processor_handlers = [result_recorder]
        self._add_result_printer(result_recorder, result_processor_handlers)
        result_processor = ResultProcessor(
            result_queue, result_processor_handlers
        )
        command_result_recorder = CommandResultRecorder(
            result_queue, result_recorder, result_processor
        )

        return S3TransferHandler(
            transfer_manager, self._cli_params, command_result_recorder
        )

    def _add_result_printer(self, result_recorder, result_processor_handlers):
        if self._cli_params.get('quiet'):
            return
        elif self._cli_params.get('only_show_errors'):
            result_printer = OnlyShowErrorsResultPrinter(result_recorder)
        elif self._cli_params.get('is_stream'):
            result_printer = OnlyShowErrorsResultPrinter(result_recorder)
        elif not self._cli_params.get('progress'):
            result_printer = NoProgressResultPrinter(result_recorder)
        else:
            result_printer = ResultPrinter(
                result_recorder,
                frequency=self._cli_params.get('progress_frequency'),
                oneline=not self._cli_params.get('progress_multiline'),
            )
        result_processor_handlers.append(result_printer)


class S3TransferHandler:
    def __init__(self, transfer_manager, cli_params, result_command_recorder):
        """Backend for performing S3 transfers

        :type transfer_manager: s3transfer.manager.TransferManager
        :param transfer_manager: Transfer manager to use for transfers

        :type cli_params: dict
        :param cli_params: The parameters passed to the CLI command in the
            form of a dictionary

        :type result_command_recorder: ResultCommandRecorder
        :param result_command_recorder: The result command recorder to be
            used to get the final result of the transfer
        """
        self._transfer_manager = transfer_manager
        # TODO: Ideally the s3 transfer handler should not need to know
        # about the result command recorder. It really only needs an interface
        # for adding results to the queue. When all of the commands have
        # converted to use this transfer handler, an effort should be made
        # to replace the passing of a result command recorder with an
        # abstraction to enqueue results.
        self._result_command_recorder = result_command_recorder

        submitter_args = (
            self._transfer_manager,
            self._result_command_recorder.result_queue,
            cli_params,
        )
        self._submitters = [
            UploadStreamRequestSubmitter(*submitter_args),
            DownloadStreamRequestSubmitter(*submitter_args),
            UploadRequestSubmitter(*submitter_args),
            DownloadRequestSubmitter(*submitter_args),
            CopyRequestSubmitter(*submitter_args),
            DeleteRequestSubmitter(*submitter_args),
            LocalDeleteRequestSubmitter(*submitter_args),
        ]

    def call(self, fileinfos):
        """Process iterable of FileInfos for transfer

        :type fileinfos: iterable of FileInfos
        param fileinfos: Set of FileInfos to submit to underlying transfer
            request submitters to make transfer API calls to S3

        :rtype: CommandResult
        :returns: The result of the command that specifies the number of
            failures and warnings encountered.
        """
        with self._result_command_recorder:
            with self._transfer_manager:
                total_submissions = 0
                for fileinfo in fileinfos:
                    for submitter in self._submitters:
                        if submitter.can_submit(fileinfo):
                            if submitter.submit(fileinfo):
                                total_submissions += 1
                            break
                self._result_command_recorder.notify_total_submissions(
                    total_submissions
                )
        return self._result_command_recorder.get_command_result()


class BaseTransferRequestSubmitter:
    REQUEST_MAPPER_METHOD = None

    def __init__(self, transfer_manager, result_queue, cli_params):
        """Submits transfer requests to the TransferManager

        Given a FileInfo object and provided CLI parameters, it will add the
        necessary extra arguments and subscribers in making a call to the
        TransferManager.

        :type transfer_manager: s3transfer.manager.TransferManager
        :param transfer_manager: The underlying transfer manager

        :type result_queue: queue.Queue
        :param result_queue: The result queue to use

        :type cli_params: dict
        :param cli_params: The associated CLI parameters passed in to the
            command as a dictionary.
        """
        self._transfer_manager = transfer_manager
        self._result_queue = result_queue
        self._cli_params = cli_params

    def submit(self, fileinfo):
        """Submits a transfer request based on the FileInfo provided

        There is no guarantee that the transfer request will be made on
        behalf of the fileinfo as a fileinfo may be skipped based on
        circumstances in which the transfer is not possible.

        :type fileinfo: awscli.customizations.s3.fileinfo.FileInfo
        :param fileinfo: The FileInfo to be used to submit a transfer
            request to the underlying transfer manager.

        :rtype: s3transfer.futures.TransferFuture
        :returns: A TransferFuture representing the transfer if it the
            transfer was submitted. If it was not submitted nothing
            is returned.
        """
        should_skip = self._warn_and_signal_if_skip(fileinfo)
        if not should_skip:
            return self._do_submit(fileinfo)

    def can_submit(self, fileinfo):
        """Checks whether it can submit a particular FileInfo

        :type fileinfo: awscli.customizations.s3.fileinfo.FileInfo
        :param fileinfo: The FileInfo to check if the transfer request
            submitter can handle.

        :returns: True if it can use the provided FileInfo to make a transfer
            request to the underlying transfer manager. False, otherwise.
        """
        raise NotImplementedError('can_submit()')

    def _do_submit(self, fileinfo):
        extra_args = {}
        if self.REQUEST_MAPPER_METHOD:
            self.REQUEST_MAPPER_METHOD(extra_args, self._cli_params)
        if not self._cli_params.get('dryrun'):
            return self._submit_transfer_request(
                fileinfo, extra_args, self._get_subscribers(fileinfo)
            )
        else:
            self._submit_dryrun(fileinfo)

    def _get_subscribers(self, fileinfo):
        subscribers = []
        result_subscriber_kwargs = self._get_result_subscriber_kwargs(fileinfo)
        self._add_provide_size_subscriber(subscribers, fileinfo)
        subscribers.append(ProvideETagSubscriber(fileinfo.etag))
        subscribers.append(QueuedResultSubscriber(**result_subscriber_kwargs))
        self._add_additional_subscribers(subscribers, fileinfo)
        subscribers.extend(
            [
                ProgressResultSubscriber(**result_subscriber_kwargs),
                DoneResultSubscriber(**result_subscriber_kwargs),
            ]
        )
        return subscribers

    def _get_result_subscriber_kwargs(self, fileinfo):
        src, dest = self._format_src_dest(fileinfo)
        return {
            'result_queue': self._result_queue,
            'transfer_type': self._get_transfer_type(fileinfo),
            'src': src,
            'dest': dest,
            'size': fileinfo.size,
        }

    def _submit_dryrun(self, fileinfo):
        transfer_type = self._get_transfer_type(fileinfo)
        src, dest = self._format_src_dest(fileinfo)
        self._result_queue.put(
            DryRunResult(transfer_type=transfer_type, src=src, dest=dest)
        )

    def _add_provide_size_subscriber(self, subscribers, fileinfo):
        subscribers.append(ProvideSizeSubscriber(fileinfo.size))

    def _add_additional_subscribers(self, subscribers, fileinfo):
        pass

    def _submit_transfer_request(self, fileinfo, extra_args, subscribers):
        raise NotImplementedError('_submit_transfer_request()')

    def _warn_and_signal_if_skip(self, fileinfo):
        for warning_handler in self._get_warning_handlers():
            if warning_handler(fileinfo):
                # On the first warning handler that returns a signal to skip
                # immediately propogate this signal and no longer check
                # the other warning handlers as no matter what the file will
                # be skipped.
                return True

    def _get_warning_handlers(self):
        # Returns a list of warning handlers, which are callables that
        # take in a single parameter representing a FileInfo. It will then
        # add a warning to result_queue if needed and return True if
        # that FileInfo should be skipped.
        return []

    def _should_inject_content_type(self):
        return self._cli_params.get(
            'guess_mime_type'
        ) and not self._cli_params.get('content_type')

    def _warn_glacier(self, fileinfo):
        if not self._cli_params.get('force_glacier_transfer'):
            if not fileinfo.is_glacier_compatible():
                LOGGER.debug(
                    'Encountered glacier object s3://%s. Not performing '
                    '%s on object.' % (fileinfo.src, fileinfo.operation_name)
                )
                if not self._cli_params.get('ignore_glacier_warnings'):
                    warning = create_warning(
                        's3://' + fileinfo.src,
                        'Object is of storage class GLACIER. Unable to '
                        'perform %s operations on GLACIER objects. You must '
                        'restore the object to be able to perform the '
                        'operation. See aws s3 %s help for additional '
                        'parameter options to ignore or force these '
                        'transfers.'
                        % (fileinfo.operation_name, fileinfo.operation_name),
                    )
                    self._result_queue.put(warning)
                return True
        return False

    def _warn_parent_reference(self, fileinfo):
        # normpath() will use the OS path separator so we
        # need to take that into account when checking for a parent prefix.
        parent_prefix = '..' + os.path.sep
        escapes_cwd = os.path.normpath(fileinfo.compare_key).startswith(
            parent_prefix
        )
        if escapes_cwd:
            warning = create_warning(
                fileinfo.compare_key, "File references a parent directory."
            )
            self._result_queue.put(warning)
            return True
        return False

    def _get_transfer_type(self, fileinfo):
        if self._cli_params.get('is_move', False):
            return 'move'
        return fileinfo.operation_name

    def _format_src_dest(self, fileinfo):
        """Returns formatted versions of a fileinfos source and destination."""
        raise NotImplementedError('_format_src_dest')

    def _format_local_path(self, path):
        return relative_path(path)

    def _format_s3_path(self, path):
        if path.startswith('s3://'):
            return path
        return 's3://' + path


class UploadRequestSubmitter(BaseTransferRequestSubmitter):
    REQUEST_MAPPER_METHOD = RequestParamsMapper.map_put_object_params

    def can_submit(self, fileinfo):
        return fileinfo.operation_name == 'upload'

    def _add_additional_subscribers(self, subscribers, fileinfo):
        if self._should_inject_content_type():
            subscribers.append(ProvideUploadContentTypeSubscriber())
        if self._cli_params.get('is_move', False):
            subscribers.append(DeleteSourceFileSubscriber())

    def _submit_transfer_request(self, fileinfo, extra_args, subscribers):
        bucket, key = find_bucket_key(fileinfo.dest)
        filein = self._get_filein(fileinfo)
        return self._transfer_manager.upload(
            fileobj=filein,
            bucket=bucket,
            key=key,
            extra_args=extra_args,
            subscribers=subscribers,
        )

    def _get_filein(self, fileinfo):
        return fileinfo.src

    def _get_warning_handlers(self):
        return [self._warn_if_too_large]

    def _warn_if_too_large(self, fileinfo):
        if getattr(fileinfo, 'size') and fileinfo.size > MAX_UPLOAD_SIZE:
            file_path = relative_path(fileinfo.src)
            warning_message = "File %s exceeds s3 upload limit of %s." % (
                file_path,
                human_readable_size(MAX_UPLOAD_SIZE),
            )
            warning = create_warning(
                file_path, warning_message, skip_file=False
            )
            self._result_queue.put(warning)

    def _format_src_dest(self, fileinfo):
        src = self._format_local_path(fileinfo.src)
        dest = self._format_s3_path(fileinfo.dest)
        return src, dest


class DownloadRequestSubmitter(BaseTransferRequestSubmitter):
    REQUEST_MAPPER_METHOD = RequestParamsMapper.map_get_object_params

    def can_submit(self, fileinfo):
        return fileinfo.operation_name == 'download'

    def _add_additional_subscribers(self, subscribers, fileinfo):
        subscribers.append(DirectoryCreatorSubscriber())
        subscribers.append(
            ProvideLastModifiedTimeSubscriber(
                fileinfo.last_update, self._result_queue
            )
        )
        if self._cli_params.get('is_move', False):
            subscribers.append(
                DeleteSourceObjectSubscriber(fileinfo.source_client)
            )
        if fileinfo.case_conflict_submitted is not None:
            subscribers.append(
                CaseConflictCleanupSubscriber(
                    fileinfo.case_conflict_submitted,
                    fileinfo.case_conflict_key,
                )
            )

    def _submit_transfer_request(self, fileinfo, extra_args, subscribers):
        bucket, key = find_bucket_key(fileinfo.src)
        fileout = self._get_fileout(fileinfo)
        return self._transfer_manager.download(
            fileobj=fileout,
            bucket=bucket,
            key=key,
            extra_args=extra_args,
            subscribers=subscribers,
        )

    def _get_fileout(self, fileinfo):
        return fileinfo.dest

    def _get_warning_handlers(self):
        return [
            self._warn_glacier,
            self._warn_parent_reference,
            self._warn_if_file_exists_with_no_overwrite,
        ]

    def _warn_if_file_exists_with_no_overwrite(self, fileinfo):
        """
        Warning handler to skip downloads when no-overwrite is set and local file exists.

        This method prevents overwriting existing local files during S3 download operations
        when the --no-overwrite flag is specified. It checks if the destination file already
        exists on the local filesystem and skips the download if found.

        :type fileinfo: FileInfo
        :param fileinfo: The FileInfo object containing transfer details

        :rtype: bool
        :returns: True if the file should be skipped (exists and no-overwrite is set),
                False if the download should proceed
        """
        if not self._cli_params.get('no_overwrite'):
            return False
        fileout = self._get_fileout(fileinfo)
        if os.path.exists(fileout):
            LOGGER.debug(
                f"warning: skipping {fileinfo.src} -> {fileinfo.dest}, file exists at destination"
            )
            return True
        return False

    def _format_src_dest(self, fileinfo):
        src = self._format_s3_path(fileinfo.src)
        dest = self._format_local_path(fileinfo.dest)
        return src, dest


class CopyRequestSubmitter(BaseTransferRequestSubmitter):
    REQUEST_MAPPER_METHOD = RequestParamsMapper.map_copy_object_params

    def can_submit(self, fileinfo):
        return fileinfo.operation_name == 'copy'

    def _add_additional_subscribers(self, subscribers, fileinfo):
        if not self._cli_params.get('metadata_directive'):
            self._add_copy_props_subscribers(subscribers, fileinfo)
        if self._cli_params.get('is_move', False):
            subscribers.append(
                DeleteCopySourceObjectSubscriber(fileinfo.source_client)
            )

    def _add_copy_props_subscribers(self, subscribers, fileinfo):
        copy_props_factory = CopyPropsSubscriberFactory(
            self._transfer_manager.client,
            self._transfer_manager.config,
            self._cli_params,
        )
        subscribers.extend(copy_props_factory.get_subscribers(fileinfo))

    def _submit_transfer_request(self, fileinfo, extra_args, subscribers):
        bucket, key = find_bucket_key(fileinfo.dest)
        source_bucket, source_key = find_bucket_key(fileinfo.src)
        copy_source = {'Bucket': source_bucket, 'Key': source_key}
        return self._transfer_manager.copy(
            bucket=bucket,
            key=key,
            copy_source=copy_source,
            extra_args=extra_args,
            subscribers=subscribers,
            source_client=fileinfo.source_client,
        )

    def _get_warning_handlers(self):
        return [
            self._warn_glacier,
        ]

    def _format_src_dest(self, fileinfo):
        src = self._format_s3_path(fileinfo.src)
        dest = self._format_s3_path(fileinfo.dest)
        return src, dest


class UploadStreamRequestSubmitter(UploadRequestSubmitter):
    def can_submit(self, fileinfo):
        return fileinfo.operation_name == 'upload' and self._cli_params.get(
            'is_stream'
        )

    def _add_provide_size_subscriber(self, subscribers, fileinfo):
        expected_size = self._cli_params.get('expected_size', None)
        if expected_size is not None:
            subscribers.append(ProvideSizeSubscriber(int(expected_size)))

    def _add_additional_subscribers(self, subscribers, fileinfo):
        pass

    def _get_filein(self, fileinfo):
        binary_stdin = get_binary_stdin()
        return NonSeekableStream(binary_stdin)

    def _format_local_path(self, path):
        return '-'


class DownloadStreamRequestSubmitter(DownloadRequestSubmitter):
    def can_submit(self, fileinfo):
        return fileinfo.operation_name == 'download' and self._cli_params.get(
            'is_stream'
        )

    def _add_provide_size_subscriber(self, subscribers, fileinfo):
        pass

    def _add_additional_subscribers(self, subscribers, fileinfo):
        pass

    def _get_fileout(self, fileinfo):
        return StdoutBytesWriter()

    def _format_local_path(self, path):
        return '-'


class DeleteRequestSubmitter(BaseTransferRequestSubmitter):
    REQUEST_MAPPER_METHOD = RequestParamsMapper.map_delete_object_params

    def can_submit(self, fileinfo):
        return (
            fileinfo.operation_name == 'delete' and fileinfo.src_type == 's3'
        )

    def _add_provide_size_subscriber(self, subscribers, fileinfo):
        pass

    def _submit_transfer_request(self, fileinfo, extra_args, subscribers):
        bucket, key = find_bucket_key(fileinfo.src)
        return self._transfer_manager.delete(
            bucket=bucket,
            key=key,
            extra_args=extra_args,
            subscribers=subscribers,
        )

    def _format_src_dest(self, fileinfo):
        return self._format_s3_path(fileinfo.src), None


class LocalDeleteRequestSubmitter(BaseTransferRequestSubmitter):
    REQUEST_MAPPER_METHOD = None

    def can_submit(self, fileinfo):
        return (
            fileinfo.operation_name == 'delete'
            and fileinfo.src_type == 'local'
        )

    def _submit_transfer_request(self, fileinfo, extra_args, subscribers):
        # This is quirky but essentially instead of relying on a built-in
        # method of s3 transfer, the logic lives directly in the submitter.
        # The reason a explicit delete local file does not
        # live in s3transfer is because it is outside the scope of s3transfer;
        # it should only have interfaces for interacting with S3. Therefore,
        # the burden of this functionality should live in the CLI.

        # The main downsides in doing this is that delete and the result
        # creation happens in the main thread as opposed to a separate thread
        # in s3transfer. However, this is not too big of a downside because
        # deleting a local file only happens for sync --delete downloads and
        # is very fast compared to all of the other types of transfers.
        src, dest = self._format_src_dest(fileinfo)
        result_kwargs = {'transfer_type': 'delete', 'src': src, 'dest': dest}
        try:
            self._result_queue.put(
                QueuedResult(total_transfer_size=0, **result_kwargs)
            )
            os.remove(fileinfo.src)
            self._result_queue.put(SuccessResult(**result_kwargs))
        except Exception as e:
            self._result_queue.put(FailureResult(exception=e, **result_kwargs))
        return True

    def _format_src_dest(self, fileinfo):
        return self._format_local_path(fileinfo.src), None
